import asyncio
import logging
from collections.abc import Mapping, Sequence
from contextlib import AsyncExitStack
from dataclasses import dataclass
from datetime import datetime
from typing import Any, Final

from ai.backend.common.clients.valkey_client.valkey_schedule import ValkeyScheduleClient
from ai.backend.common.events.dispatcher import EventProducer
from ai.backend.common.events.event_types.kernel.anycast import (
    KernelCancelledAnycastEvent,
    KernelCreatingAnycastEvent,
    KernelPreparingAnycastEvent,
    KernelPullingAnycastEvent,
    KernelStartedAnycastEvent,
    KernelTerminatedAnycastEvent,
)
from ai.backend.common.events.event_types.schedule.anycast import (
    DoSokovanProcessIfNeededEvent,
    DoSokovanProcessScheduleEvent,
)
from ai.backend.common.events.event_types.session.broadcast import (
    SchedulingBroadcastEvent,
)
from ai.backend.common.events.types import AbstractBroadcastEvent
from ai.backend.common.leader.tasks import EventTaskSpec
from ai.backend.common.types import AccessKey, AgentId, SessionId
from ai.backend.logging import BraceStyleAdapter
from ai.backend.manager.config.provider import ManagerConfigProvider
from ai.backend.manager.data.kernel.types import KernelStatus
from ai.backend.manager.data.session.types import (
    SchedulingResult,
    SessionStatus,
    StatusTransitions,
    TransitionStatus,
)
from ai.backend.manager.defs import SERVICE_MAX_RETRIES
from ai.backend.manager.metrics.scheduler import SchedulerOperationMetricObserver
from ai.backend.manager.repositories.base import BatchQuerier
from ai.backend.manager.repositories.base.creator import BulkCreator
from ai.backend.manager.repositories.base.pagination import NoPagination, OffsetPagination
from ai.backend.manager.repositories.base.updater import BatchUpdater
from ai.backend.manager.repositories.scheduler.options import KernelConditions, SessionConditions
from ai.backend.manager.repositories.scheduler.repository import SchedulerRepository
from ai.backend.manager.repositories.scheduler.updaters import SessionStatusBatchUpdaterSpec
from ai.backend.manager.repositories.scheduling_history.creators import (
    SessionSchedulingHistoryCreatorSpec,
)
from ai.backend.manager.scheduler.types import ScheduleType
from ai.backend.manager.sokovan.data import (
    KernelCreationInfo,
    PromotionSpec,
    SessionWithKernels,
)
from ai.backend.manager.sokovan.recorder.types import ExecutionRecord
from ai.backend.manager.sokovan.recorder.utils import extract_sub_steps_for_entity
from ai.backend.manager.sokovan.scheduler.scheduler import SchedulerComponents
from ai.backend.manager.sokovan.scheduling_controller import SchedulingController
from ai.backend.manager.types import DistributedLockFactory

from .factory import CoordinatorHandlers
from .handlers import SessionLifecycleHandler
from .handlers.kernel import KernelLifecycleHandler
from .handlers.observer import KernelObserver
from .hooks.registry import HookRegistry
from .kernel import KernelStateEngine
from .post_processors import (
    KernelPostProcessor,
    KernelPostProcessorContext,
    PostProcessor,
    PostProcessorContext,
    create_kernel_post_processors,
    create_session_post_processors,
)
from .recorder import SessionRecorderContext
from .results import (
    KernelExecutionResult,
    KernelStatusTransitions,
    SessionExecutionResult,
    SessionTransitionInfo,
)

log = BraceStyleAdapter(logging.getLogger(__name__))

# Status timeout thresholds (in seconds) for failure classification
# Sessions exceeding these times in a status are classified as 'expired'
STATUS_TIMEOUT_MAP: dict[SessionStatus, float] = {
    SessionStatus.PREPARING: 900.0,  # 15 minutes
    SessionStatus.PULLING: 900.0,  # 15 minutes
    SessionStatus.CREATING: 600.0,  # 10 minutes
}

# Batch size for observer kernel processing
_OBSERVER_BATCH_SIZE: Final[int] = 500


@dataclass
class SchedulerTaskSpec:
    """Specification for a scheduler's periodic task."""

    schedule_type: ScheduleType
    short_interval: float | None = None  # None means no short-cycle task
    long_interval: float = 60.0
    initial_delay: float = 30.0

    def create_if_needed_event(self) -> DoSokovanProcessIfNeededEvent:
        """Create event for checking if processing is needed."""
        return DoSokovanProcessIfNeededEvent(self.schedule_type.value)

    def create_process_event(self) -> DoSokovanProcessScheduleEvent:
        """Create event for forced processing."""
        return DoSokovanProcessScheduleEvent(self.schedule_type.value)

    @property
    def short_task_name(self) -> str:
        """Name for the short-cycle task."""
        return f"sokovan_process_if_needed_{self.schedule_type.value}"

    @property
    def long_task_name(self) -> str:
        """Name for the long-cycle task."""
        return f"sokovan_process_schedule_{self.schedule_type.value}"


@dataclass
class HookExecutionResult:
    """Result of hook execution for sessions transitioning to a new status.

    Contains both the session transition info and full session data
    for sessions that successfully passed hooks.
    """

    successful_sessions: list[SessionTransitionInfo]
    """Sessions that passed hooks and should proceed with transition."""

    full_session_data: list[SessionWithKernels]
    """Full session+kernel data for successful sessions (for occupied_slots calculation)."""


@dataclass
class FailureClassificationResult:
    """Result of classifying failures into give_up, expired, and need_retry.

    Classification priority (first match wins):
    1. give_up: phase_attempts >= SERVICE_MAX_RETRIES
    2. expired: status_changed elapsed > STATUS_TIMEOUT_MAP threshold
    3. need_retry: default (can be retried)
    """

    give_up: list[SessionTransitionInfo]
    """Sessions that exceeded max retries - should transition to give_up status."""

    expired: list[SessionTransitionInfo]
    """Sessions that exceeded timeout threshold - should transition to expired status."""

    need_retry: list[SessionTransitionInfo]
    """Sessions that can be retried - should transition to need_retry status."""


class ScheduleCoordinator:
    """
    Coordinate scheduling operations based on scheduling needs.
    Handles the actual scheduling logic and state management.
    """

    _valkey_schedule: ValkeyScheduleClient
    _components: SchedulerComponents
    _handlers: CoordinatorHandlers
    _scheduling_controller: SchedulingController
    _repository: SchedulerRepository
    _operation_metrics: SchedulerOperationMetricObserver
    _kernel_state_engine: KernelStateEngine
    _lock_factory: DistributedLockFactory
    _config_provider: ManagerConfigProvider
    _event_producer: EventProducer
    _hook_registry: HookRegistry
    _post_processors: Sequence[PostProcessor]
    _kernel_post_processors: Sequence[KernelPostProcessor]

    def __init__(
        self,
        valkey_schedule: ValkeyScheduleClient,
        components: SchedulerComponents,
        handlers: CoordinatorHandlers,
        scheduling_controller: SchedulingController,
        event_producer: EventProducer,
        lock_factory: DistributedLockFactory,
    ) -> None:
        self._valkey_schedule = valkey_schedule
        self._components = components
        self._handlers = handlers
        self._scheduling_controller = scheduling_controller
        self._repository = components.repository
        self._event_producer = event_producer
        self._lock_factory = lock_factory
        self._config_provider = components.config_provider
        self._operation_metrics = SchedulerOperationMetricObserver.instance()

        # Initialize kernel state engine with the component's repository
        self._kernel_state_engine = KernelStateEngine(components.repository)

        # Initialize hook registry from components
        self._hook_registry = components.hook_registry

        # Initialize post-processors for session handlers
        self._post_processors = create_session_post_processors(
            scheduling_controller,
            components.repository,
        )

        # Initialize post-processors for kernel handlers
        self._kernel_post_processors = create_kernel_post_processors(
            scheduling_controller,
        )

    async def process_lifecycle_schedule(
        self,
        schedule_type: ScheduleType,
    ) -> bool:
        """Process a lifecycle schedule type using the DeploymentCoordinator pattern.

        This method processes each scaling group independently:
        1. Iterates over all schedulable scaling groups
        2. For each scaling group:
           - Creates a SessionRecorderContext for the scaling group
           - Queries sessions based on handler's target_statuses() and target_kernel_statuses()
           - Executes handler logic
           - Applies status transitions immediately
           - Emits metrics per scaling group
           - Runs post-processing per scaling group

        This per-scaling-group approach prevents accumulating too many sessions
        in memory and ensures status updates are applied promptly.

        Args:
            schedule_type: Type of scheduling operation

        Returns:
            True if operation was performed, False otherwise
        """
        # Check kernel observers first (no state transitions)
        kernel_observer = self._handlers.kernel_observers.get(schedule_type)
        if kernel_observer:
            return await self._process_observer_schedule(schedule_type, kernel_observer)

        # Check promotion specs, then kernel handlers, then lifecycle handlers
        promotion_spec = self._handlers.promotion_specs.get(schedule_type)
        if promotion_spec:
            return await self._process_promotion_schedule(schedule_type, promotion_spec)

        kernel_handler = self._handlers.kernel_handlers.get(schedule_type)
        if kernel_handler:
            return await self._process_kernel_schedule(schedule_type, kernel_handler)

        lifecycle_handler = self._handlers.lifecycle_handlers.get(schedule_type)
        if lifecycle_handler:
            return await self._process_lifecycle_handler_schedule(schedule_type, lifecycle_handler)

        log.warning("No handler for schedule type: {}", schedule_type.value)
        return False

    async def _process_lifecycle_handler_schedule(
        self,
        schedule_type: ScheduleType,
        handler: SessionLifecycleHandler,
    ) -> bool:
        """Process a lifecycle handler schedule type."""
        try:
            log.debug("Processing lifecycle schedule type: {}", schedule_type.value)

            async with AsyncExitStack() as stack:
                stack.enter_context(self._operation_metrics.measure_operation(handler.name()))

                # Acquire lock if needed
                if handler.lock_id is not None:
                    lock_lifetime = (
                        self._config_provider.config.manager.session_schedule_lock_lifetime
                    )
                    await stack.enter_async_context(
                        self._lock_factory(handler.lock_id, lock_lifetime)
                    )

                # Process each scaling group in parallel
                scaling_groups = await self._repository.get_schedulable_scaling_groups()

                results = await asyncio.gather(
                    *[
                        self._process_scaling_group(handler, schedule_type, scaling_group)
                        for scaling_group in scaling_groups
                    ],
                    return_exceptions=True,
                )

                # Log any exceptions that occurred during parallel processing
                for scaling_group, result in zip(scaling_groups, results, strict=True):
                    if isinstance(result, BaseException):
                        log.error(
                            "Error processing scaling group {} for {}: {}",
                            scaling_group,
                            schedule_type.value,
                            result,
                        )

            return True

        except Exception as e:
            log.exception(
                "Error processing lifecycle schedule type {}: {}",
                schedule_type.value,
                e,
            )
            raise

    async def _process_promotion_schedule(
        self,
        schedule_type: ScheduleType,
        spec: PromotionSpec,
    ) -> bool:
        """Process a promotion spec schedule type.

        Promotion specs define query conditions and target status declaratively.
        The Coordinator processes sessions matching the spec directly.
        """
        try:
            log.debug("Processing promotion schedule type: {}", schedule_type.value)

            with self._operation_metrics.measure_operation(spec.name):
                # Process each scaling group in parallel
                scaling_groups = await self._repository.get_schedulable_scaling_groups()

                results = await asyncio.gather(
                    *[
                        self._process_promotion_scaling_group(spec, schedule_type, scaling_group)
                        for scaling_group in scaling_groups
                    ],
                    return_exceptions=True,
                )

                # Log any exceptions that occurred during parallel processing
                for scaling_group, result in zip(scaling_groups, results, strict=True):
                    if isinstance(result, BaseException):
                        log.error(
                            "Error processing scaling group {} for {}: {}",
                            scaling_group,
                            schedule_type.value,
                            result,
                        )

            return True

        except Exception as e:
            log.exception(
                "Error processing promotion schedule type {}: {}",
                schedule_type.value,
                e,
            )
            raise

    async def _process_kernel_schedule(
        self,
        schedule_type: ScheduleType,
        handler: KernelLifecycleHandler,
    ) -> bool:
        """Process a kernel handler schedule type.

        Kernel handlers operate on kernels directly. The flow:
        1. Query sessions based on handler's target_kernel_statuses
        2. Extract kernels from sessions
        3. Execute handler logic
        4. Apply kernel status transitions based on result
        """
        try:
            log.debug("Processing kernel schedule type: {}", schedule_type.value)

            async with AsyncExitStack() as stack:
                stack.enter_context(self._operation_metrics.measure_operation(handler.name()))

                # Acquire lock if needed
                if handler.lock_id is not None:
                    lock_lifetime = (
                        self._config_provider.config.manager.session_schedule_lock_lifetime
                    )
                    await stack.enter_async_context(
                        self._lock_factory(handler.lock_id, lock_lifetime)
                    )

                # Process each scaling group in parallel
                scaling_groups = await self._repository.get_schedulable_scaling_groups()

                results = await asyncio.gather(
                    *[
                        self._process_kernel_scaling_group(handler, schedule_type, scaling_group)
                        for scaling_group in scaling_groups
                    ],
                    return_exceptions=True,
                )

                # Log any exceptions that occurred during parallel processing
                for scaling_group, result in zip(scaling_groups, results, strict=True):
                    if isinstance(result, BaseException):
                        log.error(
                            "Error processing scaling group {} for {}: {}",
                            scaling_group,
                            schedule_type.value,
                            result,
                        )

            return True

        except Exception as e:
            log.exception(
                "Error processing kernel schedule type {}: {}",
                schedule_type.value,
                e,
            )
            raise

    async def _process_observer_schedule(
        self,
        schedule_type: ScheduleType,
        observer: KernelObserver,
    ) -> bool:
        """Process a kernel observer schedule type.

        Observers do not change kernel state - they only collect data.
        Used for fair share usage tracking, metrics collection, etc.

        Args:
            schedule_type: Type of scheduling operation
            observer: The kernel observer to execute

        Returns:
            True if operation was performed, False otherwise
        """
        try:
            log.debug(
                "[Coordinator] Processing observer schedule: type={}, observer={}",
                schedule_type.value,
                observer.name(),
            )

            async with AsyncExitStack() as stack:
                stack.enter_context(self._operation_metrics.measure_operation(observer.name()))

                # Process each scaling group in parallel
                scaling_groups = await self._repository.get_schedulable_scaling_groups()

                log.debug(
                    "[Coordinator] Found {} scaling groups to observe: {}",
                    len(scaling_groups),
                    scaling_groups,
                )

                results = await asyncio.gather(
                    *[
                        self._process_observer_scaling_group(observer, scaling_group)
                        for scaling_group in scaling_groups
                    ],
                    return_exceptions=True,
                )

                # Log any exceptions that occurred during parallel processing
                for scaling_group, result in zip(scaling_groups, results, strict=True):
                    if isinstance(result, BaseException):
                        log.error(
                            "Error observing scaling group {} for {}: {}",
                            scaling_group,
                            schedule_type.value,
                            result,
                        )

            log.debug("[Coordinator] Observer schedule {} completed", schedule_type.value)
            return True

        except Exception as e:
            log.exception(
                "Error processing observer schedule type {}: {}",
                schedule_type.value,
                e,
            )
            raise

    async def _process_observer_scaling_group(
        self,
        observer: KernelObserver,
        scaling_group: str,
    ) -> None:
        """Process a single scaling group for the given observer.

        This method:
        1. Queries kernels using observer's query condition
        2. Processes kernels in batches with pagination
        3. Executes observer logic (no state transitions)
        4. Emits metrics

        Args:
            observer: The kernel observer to execute
            scaling_group: The scaling group to process
        """
        log.debug(
            "[Coordinator] Processing observer {} for scaling_group={}",
            observer.name(),
            scaling_group,
        )
        condition = observer.get_query_condition(scaling_group)

        # Process in batches with pagination for large result sets
        offset = 0
        total_observed = 0

        while True:
            querier = BatchQuerier(
                pagination=OffsetPagination(limit=_OBSERVER_BATCH_SIZE, offset=offset),
                conditions=[condition],
            )

            kernel_result = await self._repository.search_kernels_for_handler(querier)

            log.debug(
                "[Coordinator] Observer {} batch: offset={}, items_count={}, has_next_page={}",
                observer.name(),
                offset,
                len(kernel_result.items),
                kernel_result.has_next_page,
            )

            if not kernel_result.items:
                log.debug("[Coordinator] Observer {} no items found, exiting loop", observer.name())
                break

            # Execute observer logic (no status transitions)
            result = await observer.observe(scaling_group, kernel_result.items)
            total_observed += result.observed_count

            # Check if there are more pages
            if not kernel_result.has_next_page:
                break

            offset += _OBSERVER_BATCH_SIZE

        log.debug(
            "[Coordinator] Observer {} completed: total_observed={}",
            observer.name(),
            total_observed,
        )

        # Emit metrics (total observed count across all batches)
        if total_observed > 0:
            self._operation_metrics.observe_success(
                operation=observer.name(),
                count=total_observed,
            )

    async def _process_kernel_scaling_group(
        self,
        handler: KernelLifecycleHandler,
        _schedule_type: ScheduleType,
        scaling_group: str,
    ) -> None:
        """Process a single scaling group for the given kernel handler.

        This method:
        1. Queries sessions to get kernels for the handler
        2. Extracts kernels matching the target status
        3. Executes handler logic
        4. Applies kernel status transitions

        Args:
            handler: The kernel handler to execute
            schedule_type: Type of scheduling operation
            scaling_group: The scaling group to process
        """
        # Build querier with kernel conditions
        target_kernel_statuses = handler.target_kernel_statuses()

        querier = BatchQuerier(
            pagination=NoPagination(),
            conditions=[
                KernelConditions.by_scaling_group(scaling_group),
                KernelConditions.by_statuses(target_kernel_statuses),
            ],
        )

        kernel_result = await self._repository.search_kernels_for_handler(querier)

        if not kernel_result.items:
            return

        # Execute handler logic with kernels
        result = await handler.execute(scaling_group, kernel_result.items)

        # Apply kernel status transitions based on handler's status_transitions
        await self._handle_kernel_result(handler, result)

        # Emit metrics per scaling group
        self._operation_metrics.observe_success(
            operation=handler.name(),
            count=result.success_count(),
        )

        # Run kernel post-processors
        if result.has_transitions():
            transitions = handler.status_transitions()
            target_statuses = self._collect_kernel_target_statuses(transitions, result)
            try:
                await self._run_kernel_post_processors(result, target_statuses)
            except Exception as e:
                log.error(
                    "Error during kernel post-processing for scaling group {}: {}",
                    scaling_group,
                    e,
                )

    async def _handle_kernel_result(
        self,
        handler: KernelLifecycleHandler,
        result: KernelExecutionResult,
    ) -> None:
        """Apply kernel status transitions based on handler result.

        Unlike session handlers which use BEP-1030 pattern with multiple outcome paths,
        kernel handlers have simpler success/failure transitions.

        Args:
            handler: The kernel handler that produced the result
            result: The execution result with successes and failures
        """
        handler_name = handler.name()
        transitions = handler.status_transitions()

        # Handle failures - apply failure transition (typically to TERMINATED)
        if result.failures and transitions.failure:
            for failure in result.failures:
                await self._kernel_state_engine.mark_kernel_terminated(
                    failure.kernel_id,
                    failure.reason or "kernel_handler_failure",
                )
            log.debug(
                "{}: Terminated {} kernels",
                handler_name,
                len(result.failures),
            )

        # Handle successes - typically no status change (success means kernel is healthy)
        # Success transition is usually None, meaning no status change needed
        if result.successes:
            log.debug(
                "{}: {} kernels processed successfully (no status change)",
                handler_name,
                len(result.successes),
            )

    async def _process_scaling_group(
        self,
        handler: SessionLifecycleHandler,
        schedule_type: ScheduleType,
        scaling_group: str,
    ) -> None:
        """Process a single scaling group for the given handler.

        This method handles all processing for one scaling group:
        - Creates a SessionRecorderContext scoped to this scaling group
        - Queries and processes sessions
        - Applies status transitions
        - Emits metrics
        - Runs post-processing

        Args:
            handler: The lifecycle handler to execute
            schedule_type: Type of scheduling operation
            scaling_group: The scaling group to process
        """
        # Query sessions for this handler in this scaling group
        sessions = await self._repository.get_sessions_for_handler(
            scaling_group,
            handler.target_statuses(),
            handler.target_kernel_statuses(),
        )

        if not sessions:
            return

        # Extract session IDs for recorder entity_ids
        session_ids = [s.session_info.identity.id for s in sessions]

        # Populate phase_attempts and phase_started_at from scheduling history for failure classification
        # Get last history records (regardless of phase), then compare phase at application level
        history_map = await self._repository.get_last_session_histories(session_ids)
        handler_name = handler.name()
        for session in sessions:
            history = history_map.get(session.session_info.identity.id)
            # Only use history data if the last history is for the current phase
            if history and history.phase == handler_name:
                session.phase_attempts = history.attempts
                session.phase_started_at = history.created_at
            else:
                session.phase_attempts = 0
                session.phase_started_at = None

        # Create recorder scoped to this scaling group
        recorder_scope = f"{schedule_type.value}:{scaling_group}"
        with SessionRecorderContext.scope(recorder_scope, entity_ids=session_ids) as pool:
            # Execute handler logic
            result = await handler.execute(scaling_group, sessions)

            # Get recorded steps for history
            all_records = pool.build_all_records()

            # Apply status transitions immediately for this scaling group (BEP-1030)
            classified = await self._handle_result(handler, result, all_records, sessions)

            # Emit metrics per scaling group
            self._operation_metrics.observe_success(
                operation=handler.name(),
                count=result.success_count(),
            )

            # Common post-process: mark next schedule and invalidate cache
            if result.has_transitions():
                transitions = handler.status_transitions()
                # Collect all target statuses from success and classified failures
                target_statuses = self._collect_target_statuses(
                    transitions, result.successes, classified
                )
                try:
                    await self._run_post_processors(result, target_statuses)
                except Exception as e:
                    log.error(
                        "Error during common post-processing for scaling group {}: {}",
                        scaling_group,
                        e,
                    )

            # Log recorded steps for this scaling group
            if all_records:
                log.debug(
                    "Recorded {} sessions with execution records for {} in scaling group {}",
                    len(all_records),
                    schedule_type.value,
                    scaling_group,
                )

    async def _process_promotion_scaling_group(
        self,
        spec: PromotionSpec,
        schedule_type: ScheduleType,
        scaling_group: str,
    ) -> None:
        """Process a single scaling group for the given promotion spec.

        This method handles all processing for one scaling group using
        promotion spec semantics (ALL/ANY/NOT_ANY kernel matching).

        Args:
            spec: The promotion spec to process
            schedule_type: Type of scheduling operation
            scaling_group: The scaling group to process
        """
        querier = BatchQuerier(
            pagination=NoPagination(),
            conditions=[
                SessionConditions.by_scaling_group(scaling_group),
                SessionConditions.by_statuses(spec.target_statuses),
                SessionConditions.by_kernel_match(
                    spec.target_kernel_statuses,
                    spec.kernel_match_type,
                ),
            ],
        )

        # Query sessions (only session data, no kernels)
        session_infos = await self._repository.search_sessions_for_handler(querier)

        if not session_infos:
            return

        session_ids = [info.identity.id for info in session_infos]

        # Create recorder scoped to this scaling group
        recorder_scope = f"{schedule_type.value}:{scaling_group}"
        with SessionRecorderContext.scope(recorder_scope, entity_ids=session_ids) as pool:
            # Build transition info from matched sessions
            with SessionRecorderContext.shared_phase(
                spec.name,
                success_detail=f"Promoting to {spec.success_status.value}",
            ):
                with SessionRecorderContext.shared_step(
                    "check_kernel_status",
                    success_detail=f"All kernels ready for {spec.success_status.value}",
                ):
                    result = SessionExecutionResult()
                    for session_info in session_infos:
                        result.successes.append(
                            SessionTransitionInfo(
                                session_id=session_info.identity.id,
                                from_status=session_info.lifecycle.status,
                                reason=spec.reason,
                                creation_id=session_info.identity.creation_id,
                                access_key=AccessKey(session_info.metadata.access_key),
                            )
                        )

            # Get recorded steps for history
            all_records = pool.build_all_records()

            # Phase 2: Apply status transitions (includes hook execution)
            await self._handle_promotion_status_transitions(spec, result, all_records)

            # Emit metrics per scaling group
            self._operation_metrics.observe_success(
                operation=spec.name,
                count=result.success_count(),
            )

            # Common post-process: mark next schedule and invalidate cache
            if result.has_transitions():
                target_statuses = {spec.success_status}
                try:
                    await self._run_post_processors(result, target_statuses)
                except Exception as e:
                    log.error(
                        "Error during common post-processing for scaling group {}: {}",
                        scaling_group,
                        e,
                    )

            # Log recorded steps for this scaling group
            if all_records:
                log.debug(
                    "Recorded {} sessions with execution records for {} in scaling group {}",
                    len(all_records),
                    schedule_type.value,
                    scaling_group,
                )

    async def _handle_promotion_status_transitions(
        self,
        spec: PromotionSpec,
        result: SessionExecutionResult,
        records: Mapping[SessionId, ExecutionRecord],
    ) -> None:
        """Apply status transitions for promotion spec results.

        Promotion specs only change session status (not kernel status).
        Kernel status changes are driven by agent events, not coordinator.

        Hook execution is centralized here:
        1. Check if target status has hooks configured
        2. If hooks exist, fetch full session data and execute hooks
        3. For blocking hooks, filter out sessions where hooks failed
        4. Apply status transition for remaining sessions

        Note: Status-specific logic (e.g., occupied_slots for RUNNING) is handled
        by StatusTransitionHook in the hook registry, not here in the Coordinator.

        Args:
            spec: The promotion spec that defines the transition
            result: Execution result containing successes
            records: Mapping of session IDs to their execution records for sub_steps
        """
        if not result.successes:
            return

        # Get DB time for status transition timestamp
        current_time = await self._repository.get_db_now()

        to_status = spec.success_status
        sessions_to_transition = result.successes

        # Execute hooks if available for this status
        hook = self._hook_registry.get_hook(to_status)
        if hook:
            hook_result = await self._execute_transition_hooks(
                sessions_to_transition,
                to_status,
            )
            sessions_to_transition = hook_result.successful_sessions

        # Apply status transition for sessions that passed hooks
        if sessions_to_transition:
            # Create TransitionStatus with kernel=None (promotion doesn't change kernel status)
            transition = TransitionStatus(session=to_status, kernel=None)
            await self._apply_transition(
                spec.name,
                sessions_to_transition,
                transition,
                SchedulingResult.SUCCESS,
                records,
                current_time,
            )

            # Broadcast events for successful transitions
            await self._broadcast_transition_events(sessions_to_transition, to_status)

    async def _execute_transition_hooks(
        self,
        session_infos: list[SessionTransitionInfo],
        target_status: SessionStatus,
    ) -> HookExecutionResult:
        """Execute transition hooks for sessions moving to target_status.

        Fetches full session+kernel data, executes hooks per session,
        and returns sessions that should proceed with the transition along with
        their full session data (for occupied_slots calculation on RUNNING transition).

        All hooks are blocking - if a hook fails, the session won't transition.

        Args:
            session_infos: Sessions to execute hooks for
            target_status: The status sessions are transitioning to

        Returns:
            HookExecutionResult containing:
            - successful_sessions: SessionTransitionInfo for sessions that passed hooks
            - full_session_data: SessionWithKernels for successful sessions
        """
        if not session_infos:
            return HookExecutionResult(successful_sessions=[], full_session_data=[])

        # Fetch full session+kernel data for hook execution
        session_ids = [s.session_id for s in session_infos]
        querier = BatchQuerier(
            pagination=NoPagination(),
            conditions=[SessionConditions.by_ids(session_ids)],
        )
        full_sessions = await self._repository.search_sessions_with_kernels_for_handler(querier)

        if not full_sessions:
            log.warning(
                "No full session data found for {} sessions transitioning to {}",
                len(session_ids),
                target_status,
            )
            return HookExecutionResult(successful_sessions=[], full_session_data=[])

        # Build session_id -> SessionTransitionInfo mapping for results
        info_map = {s.session_id: s for s in session_infos}

        # Execute hooks concurrently
        hook_coroutines = [
            self._execute_single_hook(session, target_status) for session in full_sessions
        ]
        hook_results = await asyncio.gather(*hook_coroutines, return_exceptions=True)

        # Process results - all hooks are blocking
        successful_sessions: list[SessionTransitionInfo] = []
        successful_full_sessions: list[SessionWithKernels] = []
        for session, hook_result in zip(full_sessions, hook_results, strict=True):
            session_id = session.session_info.identity.id
            original_info = info_map.get(session_id)

            if original_info is None:
                continue

            if isinstance(hook_result, BaseException):
                log.error(
                    "Hook failed for session {} transitioning to {}: {}",
                    session_id,
                    target_status,
                    hook_result,
                )
                # Hook failed - don't include in successful sessions
                continue

            successful_sessions.append(original_info)
            successful_full_sessions.append(session)

        log.info(
            "Executed on_transition hooks for {} sessions transitioning to {} ({} succeeded)",
            len(full_sessions),
            target_status,
            len(successful_sessions),
        )

        return HookExecutionResult(
            successful_sessions=successful_sessions,
            full_session_data=successful_full_sessions,
        )

    async def _execute_single_hook(
        self,
        session: SessionWithKernels,
        status: SessionStatus,
    ) -> None:
        """Execute a single hook for a session.

        Args:
            session: Full session+kernel data
            status: The status the session is transitioning to
        """
        hook = self._hook_registry.get_hook(status)
        if hook:
            await hook.execute(session)

    async def _broadcast_transition_events(
        self,
        sessions: list[SessionTransitionInfo],
        to_status: SessionStatus,
    ) -> None:
        """Broadcast scheduling events for session status transitions.

        Creates SchedulingBroadcastEvent for each session and broadcasts them in batch.
        Uses session data from SessionTransitionInfo (session_id, creation_id, reason).

        Args:
            sessions: Sessions that transitioned successfully
            to_status: The target status sessions transitioned to
        """
        if not sessions:
            return

        events: list[AbstractBroadcastEvent] = []
        for session_info in sessions:
            if session_info.creation_id is None:
                log.warning(
                    "Skipping event broadcast for session {} - missing creation_id",
                    session_info.session_id,
                )
                continue

            events.append(
                SchedulingBroadcastEvent(
                    session_id=session_info.session_id,
                    creation_id=session_info.creation_id,
                    status_transition=str(to_status),
                    reason=session_info.reason or "triggered-by-scheduler",
                )
            )

        if events:
            await self._event_producer.broadcast_events_batch(events)
            log.debug(
                "Broadcast {} transition events for status {}",
                len(events),
                to_status,
            )

    async def _handle_result(
        self,
        handler: SessionLifecycleHandler,
        result: SessionExecutionResult,
        records: Mapping[SessionId, ExecutionRecord],
        sessions: list[SessionWithKernels],
    ) -> FailureClassificationResult | None:
        """Apply status transitions using handler.status_transitions() (BEP-1030).

        Handler reports what happened (successes/failures/skipped), and Coordinator applies
        policy-based classification for failures to determine the outcome:
        - need_retry: Can retry (default for failures)
        - expired: Timeout exceeded
        - give_up: Max retries exceeded

        Skipped sessions are recorded in history without status change.

        Args:
            handler: The lifecycle handler that produced the result
            result: Execution result containing successes, failures, and skipped
            records: Mapping of session IDs to their execution records for sub_steps
            sessions: Original sessions with phase_attempts for failure classification

        Returns:
            FailureClassificationResult if there were failures, None otherwise.
            Used by caller for post-processing with correct target statuses.
        """
        transitions = handler.status_transitions()
        handler_name = handler.name()
        classified: FailureClassificationResult | None = None

        # Get DB time once for all operations in this result handling
        current_time = await self._repository.get_db_now()

        # SUCCESS transitions
        if transitions.success and result.successes:
            await self._apply_transition(
                handler_name,
                result.successes,
                transitions.success,
                SchedulingResult.SUCCESS,
                records,
                current_time,
            )
            # Broadcast events for successful transitions
            if transitions.success.session:
                await self._broadcast_transition_events(
                    result.successes, transitions.success.session
                )

        # FAILURE transitions - Coordinator classifies failures into give_up/expired/need_retry
        if result.failures:
            classified = self._classify_failures(result.failures, sessions, current_time)

            # Apply transitions for each classification
            if classified.give_up and transitions.give_up:
                await self._apply_transition(
                    handler_name,
                    classified.give_up,
                    transitions.give_up,
                    SchedulingResult.GIVE_UP,
                    records,
                    current_time,
                )

            if classified.expired and transitions.expired:
                await self._apply_transition(
                    handler_name,
                    classified.expired,
                    transitions.expired,
                    SchedulingResult.EXPIRED,
                    records,
                    current_time,
                )

            if classified.need_retry and transitions.need_retry:
                await self._apply_transition(
                    handler_name,
                    classified.need_retry,
                    transitions.need_retry,
                    SchedulingResult.NEED_RETRY,
                    records,
                    current_time,
                )

        # SKIPPED - Record history without status change
        if result.skipped:
            await self._record_skipped_history(handler_name, result.skipped, records)

        return classified

    def _classify_failures(
        self,
        failures: list[SessionTransitionInfo],
        sessions: list[SessionWithKernels],
        current_time: datetime,
    ) -> FailureClassificationResult:
        """Classify failures into give_up, expired, need_retry.

        Pure classification based on conditions only. The caller (_handle_result)
        decides whether to apply transitions based on handler's transition definitions.

        Classification priority (first match wins):
        1. give_up: phase_attempts >= SERVICE_MAX_RETRIES
        2. expired: timeout exceeded
        3. need_retry: default

        Args:
            failures: Failed session transition info
            sessions: Original sessions with phase_attempts and phase_started_at populated
            current_time: Current database time for timeout comparison

        Returns:
            FailureClassificationResult with give_up, expired, need_retry lists
        """
        session_map = {s.session_info.identity.id: s for s in sessions}

        give_up_failures: list[SessionTransitionInfo] = []
        expired_failures: list[SessionTransitionInfo] = []
        retry_failures: list[SessionTransitionInfo] = []

        for failure in failures:
            session = session_map.get(failure.session_id)
            if not session:
                # Session not found - skip (shouldn't happen)
                continue

            # 1. Check max retries exceeded → give_up
            if session.phase_attempts >= SERVICE_MAX_RETRIES:
                give_up_failures.append(failure)
                continue

            # 2. Check timeout exceeded → expired
            if session.phase_started_at:
                status = session.session_info.lifecycle.status
                timeout = STATUS_TIMEOUT_MAP.get(status)
                if timeout:
                    elapsed = (current_time - session.phase_started_at).total_seconds()
                    if elapsed > timeout:
                        expired_failures.append(failure)
                        continue

            # 3. Default → need_retry
            retry_failures.append(failure)

        return FailureClassificationResult(
            give_up=give_up_failures,
            expired=expired_failures,
            need_retry=retry_failures,
        )

    async def _apply_transition(
        self,
        handler_name: str,
        session_infos: list[Any],
        transition: TransitionStatus,
        scheduling_result: SchedulingResult,
        records: Mapping[SessionId, ExecutionRecord],
        status_changed_at: datetime,
    ) -> None:
        """Apply a single transition type to sessions (BEP-1030).

        Args:
            handler_name: Name of the handler for logging and history
            session_infos: List of SessionTransitionInfo for the sessions to update
            transition: Target status transition to apply
            scheduling_result: Result type for history recording
            records: Mapping of session IDs to their execution records
            status_changed_at: Database timestamp for status change
        """
        if not session_infos:
            return

        session_ids = [s.session_id for s in session_infos]

        # Session status update
        if transition.session:
            updater = BatchUpdater(
                spec=SessionStatusBatchUpdaterSpec(
                    to_status=transition.session,
                    status_changed_at=status_changed_at,
                ),
                conditions=[SessionConditions.by_ids(session_ids)],
            )
            history_specs = [
                SessionSchedulingHistoryCreatorSpec(
                    session_id=info.session_id,
                    phase=handler_name,
                    result=scheduling_result,
                    message=f"{handler_name} {scheduling_result.value.lower()}",
                    from_status=info.from_status,
                    to_status=transition.session,
                    error_code=info.error_code,
                    sub_steps=extract_sub_steps_for_entity(info.session_id, records),
                )
                for info in session_infos
            ]
            updated = await self._repository.update_with_history(
                updater, BulkCreator(specs=history_specs)
            )
            log.debug(
                "{}: Updated {} sessions to {} ({})",
                handler_name,
                updated,
                transition.session,
                scheduling_result.value,
            )

        # Kernel status reset if transitioning to PENDING
        if transition.kernel == KernelStatus.PENDING:
            await self._apply_kernel_pending_resets(handler_name, session_ids)

    async def _apply_kernel_pending_resets(
        self,
        handler_name: str,
        session_ids: list[SessionId],
    ) -> None:
        """Reset kernels to PENDING for sessions going back to PENDING.

        When sessions exceed max retries, they go back to PENDING for re-scheduling.
        This also resets their kernels to PENDING and clears agent assignments.

        Args:
            handler_name: Name of the handler for logging
            session_ids: List of session IDs whose kernels should be reset
        """
        if not session_ids:
            return

        reset_count = await self._kernel_state_engine.reset_kernels_to_pending_for_sessions(
            session_ids,
            reason="EXCEEDED_MAX_RETRIES",
        )
        log.debug(
            "{}: Reset {} kernels to PENDING for {} sessions",
            handler_name,
            reset_count,
            len(session_ids),
        )

    async def _record_skipped_history(
        self,
        handler_name: str,
        session_infos: list[SessionTransitionInfo],
        records: Mapping[SessionId, ExecutionRecord],
    ) -> None:
        """Record history for skipped sessions without status change.

        Skipped sessions are those that were not processed due to being blocked
        by other sessions (e.g., scheduling attempt blocked by higher priority sessions).

        Args:
            handler_name: Name of the handler for history recording
            session_infos: List of skipped session information
            records: Mapping of session IDs to their execution records for sub_steps
        """
        if not session_infos:
            return

        history_specs = [
            SessionSchedulingHistoryCreatorSpec(
                session_id=info.session_id,
                phase=handler_name,
                result=SchedulingResult.SKIPPED,
                message=info.reason or f"{handler_name} skipped",
                from_status=info.from_status,
                to_status=info.from_status,  # No status change
                error_code=info.error_code,
                sub_steps=extract_sub_steps_for_entity(info.session_id, records),
            )
            for info in session_infos
        ]
        await self._repository.create_scheduling_history(BulkCreator(specs=history_specs))
        log.debug(
            "{}: Recorded {} skipped sessions in history",
            handler_name,
            len(session_infos),
        )

    def _collect_target_statuses(
        self,
        transitions: StatusTransitions,
        successes: list[SessionTransitionInfo],
        classified: FailureClassificationResult | None,
    ) -> set[SessionStatus]:
        """Collect all target statuses from successes and classified failures.

        Args:
            transitions: Status transitions defined by the handler
            successes: List of successful session transitions
            classified: Classified failures (give_up, expired, need_retry)

        Returns:
            Set of all target statuses that sessions transitioned to
        """
        target_statuses: set[SessionStatus] = set()

        # Add success target status
        if successes and transitions.success and transitions.success.session:
            target_statuses.add(transitions.success.session)

        # Add failure target statuses from classified failures
        if classified:
            if classified.give_up and transitions.give_up and transitions.give_up.session:
                target_statuses.add(transitions.give_up.session)
            if classified.expired and transitions.expired and transitions.expired.session:
                target_statuses.add(transitions.expired.session)
            if classified.need_retry and transitions.need_retry and transitions.need_retry.session:
                target_statuses.add(transitions.need_retry.session)

        return target_statuses

    async def _run_post_processors(
        self,
        result: SessionExecutionResult,
        target_statuses: set[SessionStatus],
    ) -> None:
        """Run all post-processors for session handlers in parallel.

        Args:
            result: Execution result containing successes and failures
            target_statuses: Set of target statuses sessions transitioned to
        """
        context = PostProcessorContext(result=result, target_statuses=target_statuses)
        results = await asyncio.gather(
            *[post_processor.execute(context) for post_processor in self._post_processors],
            return_exceptions=True,
        )
        for post_processor, res in zip(self._post_processors, results, strict=True):
            if isinstance(res, BaseException):
                log.warning(
                    "Post-processor {} failed: {}",
                    post_processor.__class__.__name__,
                    res,
                )

    async def _run_kernel_post_processors(
        self,
        result: KernelExecutionResult,
        target_statuses: set[KernelStatus],
    ) -> None:
        """Run all post-processors for kernel handlers in parallel.

        Args:
            result: Kernel execution result containing successes and failures
            target_statuses: Set of target kernel statuses that kernels transitioned to
        """
        context = KernelPostProcessorContext(result=result, target_statuses=target_statuses)
        results = await asyncio.gather(
            *[post_processor.execute(context) for post_processor in self._kernel_post_processors],
            return_exceptions=True,
        )
        for post_processor, res in zip(self._kernel_post_processors, results, strict=True):
            if isinstance(res, BaseException):
                log.warning(
                    "Kernel post-processor {} failed: {}",
                    post_processor.__class__.__name__,
                    res,
                )

    def _collect_kernel_target_statuses(
        self,
        transitions: KernelStatusTransitions,
        result: KernelExecutionResult,
    ) -> set[KernelStatus]:
        """Collect all target kernel statuses from successes and failures.

        Args:
            transitions: Kernel status transitions defined by the handler
            result: Kernel execution result

        Returns:
            Set of all target kernel statuses that kernels transitioned to
        """
        target_statuses: set[KernelStatus] = set()

        if result.successes and transitions.success:
            target_statuses.add(transitions.success)
        if result.failures and transitions.failure:
            target_statuses.add(transitions.failure)

        return target_statuses

    async def process_schedule(
        self,
        schedule_type: ScheduleType,
    ) -> bool:
        """
        Force processing of a specific schedule type.
        This method processes the scheduling operation even if it was not requested for guaranteed execution.
        So it should be called in long term loops.

        :param schedule_type: Type of scheduling operation
        :return: True if operation was performed, False otherwise
        """
        return await self.process_lifecycle_schedule(schedule_type)

    async def process_if_needed(self, schedule_type: ScheduleType) -> bool:
        """
        Process scheduling operation if needed (based on internal state).
        This method checks if the scheduling operation was requested and processes it if so.

        :param schedule_type: Type of scheduling operation
        :return: True if operation was performed, False otherwise
        """
        # Check internal state (uses Redis marks)
        if not await self._valkey_schedule.load_and_delete_schedule_mark(schedule_type.value):
            return False

        return await self.process_schedule(schedule_type)

    # Kernel event handling methods using the coordinator's kernel state engine

    async def handle_kernel_pulling(self, event: KernelPullingAnycastEvent) -> bool:
        """Handle kernel pulling event through the kernel state engine."""
        result = await self._kernel_state_engine.mark_kernel_pulling(event.kernel_id, event.reason)
        if result:
            # Request CHECK_PULLING_PROGRESS to monitor image pull progress
            await self._scheduling_controller.mark_scheduling_needed([
                ScheduleType.CHECK_PULLING_PROGRESS,
            ])
        return result

    async def handle_kernel_creating(self, event: KernelCreatingAnycastEvent) -> bool:
        """Handle kernel creating event through the kernel state engine."""
        return await self._kernel_state_engine.mark_kernel_creating(event.kernel_id, event.reason)

    async def handle_kernel_running(self, event: KernelStartedAnycastEvent) -> bool:
        """Handle kernel running event through the kernel state engine."""
        # Convert event data to dataclass (always present, may be empty)
        creation_info = KernelCreationInfo.from_dict(dict(event.creation_info))
        result = await self._kernel_state_engine.mark_kernel_running(
            event.kernel_id,
            event.reason,
            creation_info,
        )
        if result:
            # Request CHECK_CREATING_PROGRESS to check if session should transition to RUNNING
            await self._scheduling_controller.mark_scheduling_needed([
                ScheduleType.CHECK_CREATING_PROGRESS,
            ])
        return result

    async def handle_kernel_preparing(self, event: KernelPreparingAnycastEvent) -> bool:
        """Handle kernel preparing event through the kernel state engine."""
        result = await self._kernel_state_engine.mark_kernel_preparing(event.kernel_id)
        if result:
            # Request CHECK_PRECONDITION to check if images are ready
            await self._scheduling_controller.mark_scheduling_needed([
                ScheduleType.CHECK_PRECONDITION,
            ])
        return result

    async def handle_kernel_cancelled(self, event: KernelCancelledAnycastEvent) -> bool:
        """Handle kernel cancelled event through the kernel state engine."""
        return await self._kernel_state_engine.mark_kernel_cancelled(
            event.kernel_id, event.session_id, event.reason
        )

    async def handle_kernel_terminated(self, event: KernelTerminatedAnycastEvent) -> bool:
        """Handle kernel terminated event through the kernel state engine."""
        log.info("Handling termination of kernel {}", event.kernel_id)
        result = await self._kernel_state_engine.mark_kernel_terminated(
            event.kernel_id, event.reason, event.exit_code
        )
        if result:
            # Request DETECT_KERNEL_TERMINATION and CHECK_TERMINATING_PROGRESS
            await self._scheduling_controller.mark_scheduling_needed([
                ScheduleType.DETECT_KERNEL_TERMINATION,
                ScheduleType.CHECK_TERMINATING_PROGRESS,
            ])
        return result

    # Image-related kernel state update methods

    async def update_kernels_to_pulling_for_image(
        self,
        agent_id: AgentId,
        image: str,
        image_ref: str | None = None,
    ) -> None:
        """Update kernel status from PREPARING to PULLING for the specified image on an agent."""
        await self._kernel_state_engine.update_kernels_to_pulling_for_image(
            agent_id, image, image_ref
        )

    async def update_kernels_to_prepared_for_image(
        self,
        agent_id: AgentId,
        image: str,
        image_ref: str | None = None,
    ) -> None:
        """Update kernel status to PREPARED for the specified image on an agent."""
        result = await self._kernel_state_engine.update_kernels_to_prepared_for_image(
            agent_id, image, image_ref
        )
        if result > 0:
            log.info(
                "Updated {} kernels to PREPARED state for agent:{} image:{}",
                result,
                agent_id,
                image,
            )
            # Request scheduling to check if sessions can transition to RUNNING
            await self._scheduling_controller.mark_scheduling_needed([
                ScheduleType.CHECK_PULLING_PROGRESS,
            ])

    async def cancel_kernels_for_failed_image(
        self,
        agent_id: AgentId,
        image: str,
        error_msg: str,
        image_ref: str | None = None,
    ) -> None:
        """Cancel kernels for an image that failed to be available on an agent."""
        await self._kernel_state_engine.cancel_kernels_for_failed_image(
            agent_id, image, error_msg, image_ref
        )
        # No need to request scheduling for cancelled kernels

    @staticmethod
    def _create_task_specs() -> list[SchedulerTaskSpec]:
        """Create task specifications for all schedule types."""
        return [
            # Regular scheduling operations with both short and long cycle tasks
            SchedulerTaskSpec(
                ScheduleType.SCHEDULE,
                short_interval=2.0,
                long_interval=60.0,
                initial_delay=30.0,
            ),
            SchedulerTaskSpec(
                ScheduleType.CHECK_PRECONDITION,
                short_interval=2.0,
                long_interval=60.0,
                initial_delay=30.0,
            ),
            SchedulerTaskSpec(
                ScheduleType.START,
                short_interval=2.0,
                long_interval=60.0,
                initial_delay=30.0,
            ),
            SchedulerTaskSpec(
                ScheduleType.TERMINATE,
                short_interval=2.0,
                long_interval=60.0,
                initial_delay=30.0,
            ),
            # Sweep is a maintenance task - only needs long cycle task
            SchedulerTaskSpec(
                ScheduleType.SWEEP,
                short_interval=None,  # No short-cycle task for maintenance
                long_interval=60.0,
                initial_delay=30.0,
            ),
            # Sweep stale kernels - maintenance task to clean up kernels with stale presence
            SchedulerTaskSpec(
                ScheduleType.SWEEP_STALE_KERNELS,
                short_interval=None,  # No short-cycle task for maintenance
                long_interval=60.0,
                initial_delay=30.0,
            ),
            # Progress check operations with both short and long cycle tasks
            SchedulerTaskSpec(
                ScheduleType.CHECK_PULLING_PROGRESS,
                short_interval=2.0,
                long_interval=60.0,
                initial_delay=30.0,
            ),
            SchedulerTaskSpec(
                ScheduleType.CHECK_CREATING_PROGRESS,
                short_interval=2.0,
                long_interval=60.0,
                initial_delay=30.0,
            ),
            SchedulerTaskSpec(
                ScheduleType.CHECK_TERMINATING_PROGRESS,
                short_interval=2.0,
                long_interval=60.0,
                initial_delay=30.0,
            ),
            # Detect active sessions with any kernel TERMINATED/CANCELLED
            SchedulerTaskSpec(
                ScheduleType.DETECT_KERNEL_TERMINATION,
                short_interval=2.0,
                long_interval=60.0,
                initial_delay=30.0,
            ),
            # Fair share observation - records RUNNING kernel usage for fair share calculation
            SchedulerTaskSpec(
                ScheduleType.OBSERVE_FAIR_SHARE,
                short_interval=None,  # No short-cycle task for observation
                long_interval=300.0,  # 5 minutes
                initial_delay=60.0,  # Start 1 minute after manager starts
            ),
        ]

    def create_task_specs(self) -> list[EventTaskSpec]:
        """Create task specifications for leader-based scheduling."""
        timer_specs = self._create_task_specs()
        specs: list[EventTaskSpec] = []

        for spec in timer_specs:
            # Create short-cycle task spec if specified
            if spec.short_interval is not None:
                short_spec = EventTaskSpec(
                    name=spec.short_task_name,
                    event_factory=spec.create_if_needed_event,
                    interval=spec.short_interval,
                    initial_delay=0.0,  # Start immediately for short tasks
                )
                specs.append(short_spec)

            # Create long-cycle task spec (always present)
            long_spec = EventTaskSpec(
                name=spec.long_task_name,
                event_factory=spec.create_process_event,
                interval=spec.long_interval,
                initial_delay=spec.initial_delay,
            )
            specs.append(long_spec)

        return specs



from dataclasses import dataclass
from typing import override

from ai.backend.manager.actions.action import BaseActionResult
from ai.backend.manager.actions.types import ActionOperationType
from ai.backend.manager.data.user.types import UserData
from ai.backend.manager.models.user import UserRow
from ai.backend.manager.repositories.base.updater import Updater
from ai.backend.manager.services.user.actions.base import UserAction


@dataclass
class ModifyUserAction(UserAction):
    email: str
    updater: Updater[UserRow]

    @override
    def entity_id(self) -> str | None:
        return None

    @override
    @classmethod
    def operation_type(cls) -> ActionOperationType:
        return ActionOperationType.UPDATE


@dataclass
class ModifyUserActionResult(BaseActionResult):
    data: UserData

    @override
    def entity_id(self) -> str | None:
        return str(self.data.id)



"""Resource Slot repository package."""

from .repositories import ResourceSlotRepositories
from .repository import ResourceSlotRepository
from .upserters import AgentResourceUpserterSpec

__all__ = (
    # Repositories
    "ResourceSlotRepositories",
    "ResourceSlotRepository",
    # Upserter specs
    "AgentResourceUpserterSpec",
)



from collections.abc import Callable
from typing import Any

import sqlalchemy as sa
from sqlalchemy.orm.strategy_options import _AbstractLoad

# QueryCondition: function that takes a Select and returns a modified Select with where clause
type QueryCondition = Callable[[sa.sql.Select[Any]], sa.sql.Select[Any]]

# QueryOption: function that takes a Select and returns a modified Select with options
type QueryOption = Callable[[sa.sql.Select[Any]], sa.sql.Select[Any]]


def load_related_field(field: _AbstractLoad) -> QueryOption:
    return lambda stmt: stmt.options(field)


def join_by_related_field(field: sa.orm.attributes.InstrumentedAttribute[Any]) -> QueryOption:
    return lambda stmt: stmt.join(field)



from __future__ import annotations

from collections.abc import AsyncIterator
from contextlib import asynccontextmanager
from dataclasses import dataclass

from ai.backend.common.dependencies import DependencyComposer, DependencyStack
from ai.backend.common.etcd import AsyncEtcd
from ai.backend.manager.config.unified import ManagerUnifiedConfig
from ai.backend.manager.models.utils import ExtendedAsyncSAEngine

from .database import DatabaseDependency
from .redis import ValkeyClients, ValkeyDependency


@dataclass
class InfrastructureInput:
    """Input required for infrastructure setup.

    Contains the unified configuration and an already-initialized etcd client
    from the bootstrap stage.
    """

    config: ManagerUnifiedConfig
    etcd: AsyncEtcd


@dataclass
class InfrastructureResources:
    """Container for all initialized infrastructure resources.

    Holds initialized foundational dependencies (etcd, valkey clients, database)
    that are used throughout the server lifecycle or CLI execution.
    """

    etcd: AsyncEtcd
    valkey: ValkeyClients
    db: ExtendedAsyncSAEngine


class InfrastructureComposer(DependencyComposer[InfrastructureInput, InfrastructureResources]):
    """Composes all infrastructure dependencies as a single unit.

    Composes valkey and database dependencies in order.
    Etcd is passed from the bootstrap stage and not managed here.
    """

    @property
    def stage_name(self) -> str:
        return "infrastructure"

    @asynccontextmanager
    async def compose(
        self,
        stack: DependencyStack,
        setup_input: InfrastructureInput,
    ) -> AsyncIterator[InfrastructureResources]:
        """Compose all infrastructure dependencies.

        Args:
            stack: The dependency stack to use for composition
            setup_input: Infrastructure input containing config and etcd client

        Yields:
            InfrastructureResources containing all initialized resources
        """
        # Initialize dependency providers
        valkey_dep = ValkeyDependency()
        db_dep = DatabaseDependency()

        # Enter dependencies using the stack
        valkey = await stack.enter_dependency(
            valkey_dep,
            setup_input.config,
        )
        db = await stack.enter_dependency(
            db_dep,
            setup_input.config,
        )

        # Yield the infrastructure resources
        yield InfrastructureResources(
            etcd=setup_input.etcd,
            valkey=valkey,
            db=db,
        )



"""Revision operations action definitions."""

from .activate_revision import (
    ActivateRevisionAction,
    ActivateRevisionActionResult,
)

__all__ = [
    "ActivateRevisionAction",
    "ActivateRevisionActionResult",
]



"""CSV export stream reader for streaming responses."""

from __future__ import annotations

from collections.abc import AsyncIterator
from typing import override

from ai.backend.common.types import StreamReader

from .csv import CSVExporter


class CSVExportStreamReader(StreamReader):
    """StreamReader wrapper for CSVExporter.

    Wraps CSVExporter to implement the StreamReader interface
    for use with stream_api_handler.
    """

    _exporter: CSVExporter

    def __init__(self, exporter: CSVExporter) -> None:
        self._exporter = exporter

    @override
    async def read(self) -> AsyncIterator[bytes]:
        """Yield CSV data chunks."""
        async for chunk in self._exporter.export():
            yield chunk

    @override
    def content_type(self) -> str:
        """Return CSV content type."""
        return "text/csv; charset=utf-8"



"""fix_destory_typo_in_audit_logs_operation_column

Revision ID: 81f2dd702d93
Revises: a1b2c3d4e5f6
Create Date: 2026-01-14 10:13:17.170033

"""

import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "81f2dd702d93"
down_revision = "a1b2c3d4e5f6"
branch_labels = None
depends_on = None


def upgrade() -> None:
    # Fix typo: "destory" -> "destroy" in audit_logs.operation column
    op.execute(sa.text("UPDATE audit_logs SET operation = 'destroy' WHERE operation = 'destory'"))


def downgrade() -> None:
    # Revert: "destroy" -> "destory" in audit_logs.operation column
    op.execute(sa.text("UPDATE audit_logs SET operation = 'destory' WHERE operation = 'destroy'"))



"""Creator for RBAC scope-scoped entity insert operations."""

from __future__ import annotations

from collections.abc import Sequence
from dataclasses import dataclass, field
from typing import TypeVar

from sqlalchemy import inspect
from sqlalchemy.ext.asyncio import AsyncSession as SASession

from ai.backend.common.data.permission.types import EntityType, ScopeType
from ai.backend.manager.data.permission.id import ScopeId
from ai.backend.manager.errors.repository import UnsupportedCompositePrimaryKeyError
from ai.backend.manager.models.base import Base
from ai.backend.manager.models.rbac_models.association_scopes_entities import (
    AssociationScopesEntitiesRow,
)
from ai.backend.manager.repositories.base.creator import CreatorSpec

TRow = TypeVar("TRow", bound=Base)


# =============================================================================
# Single Entity Creator
# =============================================================================


@dataclass
class RBACEntityCreator[TRow: Base]:
    """Creator for a single entity with scope associations for RBAC.

    Creates an entity row and associates it with one or more permission scopes.
    The primary scope is required; additional scopes are optional.

    Attributes:
        spec: CreatorSpec implementation defining the row to create.
        entity_type: The entity type for RBAC association.
        scope_ref: Primary scope reference (scope_type + scope_id) for this entity.
        additional_scope_refs: Additional scope references for multi-scope entities.
    """

    spec: CreatorSpec[TRow]
    entity_type: EntityType
    scope_ref: ScopeId
    additional_scope_refs: Sequence[ScopeId] = field(default_factory=list)


@dataclass
class RBACEntityCreatorResult[TRow: Base]:
    """Result of executing a single entity creation."""

    row: TRow


async def execute_rbac_entity_creator[TRow: Base](
    db_sess: SASession,
    creator: RBACEntityCreator[TRow],
) -> RBACEntityCreatorResult[TRow]:
    """Create a scope-scoped entity with its scope association.

    Operations:
    1. Insert main entity row
    2. Flush to get DB-generated ID
    3. Extract RBAC info from spec
    4. Insert AssociationScopesEntitiesRow (scope -> entity mapping)

    The AssociationScopesEntitiesRow maps the entity to its owning scope,
    enabling scope-based entity discovery and permission inheritance.

    Args:
        db_sess: Async SQLAlchemy session (must be writable).
        creator: Creator instance with spec defining the entity to create.

    Returns:
        RBACEntityCreatorResult containing the created row.
    """
    spec = creator.spec
    row = spec.build_row()
    mapper = inspect(type(row))
    pk_columns = mapper.primary_key
    if len(pk_columns) != 1:
        raise UnsupportedCompositePrimaryKeyError(
            f"Purger only supports single-column primary keys (table: {mapper.local_table.name})",
        )

    # 1. Build and insert row
    db_sess.add(row)

    # 2. Flush to get DB-generated ID
    await db_sess.flush()

    # 3. Extract RBAC info and insert associations for all scopes
    instance_state = inspect(row)
    pk_value = instance_state.identity[0]
    all_scope_refs = [creator.scope_ref, *creator.additional_scope_refs]
    for scope_ref in all_scope_refs:
        db_sess.add(
            AssociationScopesEntitiesRow(
                scope_type=scope_ref.scope_type,
                scope_id=scope_ref.scope_id,
                entity_type=creator.entity_type,
                entity_id=str(pk_value),
            ),
        )

    return RBACEntityCreatorResult(row=row)


# =============================================================================
# Bulk Entity Creator
# =============================================================================


@dataclass
class RBACBulkEntityCreator[TRow: Base]:
    """Bulk creator for multiple entities with a single shared scope.

    Attributes:
        specs: Sequence of CreatorSpec implementations.
        scope_type: The scope type for all entities.
        scope_id: The scope ID for all entities.
        entity_type: The entity type for all entities.
    """

    specs: Sequence[CreatorSpec[TRow]]
    scope_type: ScopeType
    scope_id: str
    entity_type: EntityType


@dataclass
class RBACBulkEntityCreatorResult[TRow: Base]:
    """Result of executing a bulk entity creation."""

    rows: list[TRow]


async def execute_rbac_bulk_entity_creator[TRow: Base](
    db_sess: SASession,
    creator: RBACBulkEntityCreator[TRow],
) -> RBACBulkEntityCreatorResult[TRow]:
    """Create multiple scope-scoped entities in a single transaction.

    Operations:
    1. Build and insert all entity rows
    2. Flush to get DB-generated IDs
    3. Bulk insert AssociationScopesEntitiesRows

    Args:
        db_sess: Async SQLAlchemy session (must be writable).
        creator: Bulk creator with specs defining entities to create.

    Returns:
        RBACBulkEntityCreatorResult containing all created rows.
    """
    if not creator.specs:
        return RBACBulkEntityCreatorResult(rows=[])

    # 1. Build and add all rows
    rows = [spec.build_row() for spec in creator.specs]
    db_sess.add_all(rows)

    mapper = inspect(type(rows[0]))
    pk_columns = mapper.primary_key
    if len(pk_columns) != 1:
        raise UnsupportedCompositePrimaryKeyError(
            f"Purger only supports single-column primary keys (table: {mapper.local_table.name})",
        )

    # 2. Flush to get DB-generated IDs and insert associations
    await db_sess.flush()

    associations = [
        AssociationScopesEntitiesRow(
            scope_type=creator.scope_type,
            scope_id=creator.scope_id,
            entity_type=creator.entity_type,
            entity_id=str(inspect(row).identity[0]),
        )
        for row in rows
    ]
    db_sess.add_all(associations)

    return RBACBulkEntityCreatorResult(rows=rows)



"""Route lifecycle management module."""

from .coordinator import RouteCoordinator
from .executor import RouteExecutor
from .types import RouteExecutionError, RouteExecutionResult, RouteLifecycleType

__all__ = [
    "RouteCoordinator",
    "RouteExecutionError",
    "RouteExecutionResult",
    "RouteExecutor",
    "RouteLifecycleType",
]



"""Action for activating a deployment revision."""

from __future__ import annotations

from dataclasses import dataclass
from typing import override
from uuid import UUID

from ai.backend.manager.actions.action import BaseActionResult
from ai.backend.manager.actions.types import ActionOperationType
from ai.backend.manager.data.deployment.types import ModelDeploymentData

from .base import RevisionOperationBaseAction


@dataclass
class ActivateRevisionAction(RevisionOperationBaseAction):
    """Action to activate a specific revision to be the current revision."""

    deployment_id: UUID
    revision_id: UUID

    @override
    def entity_id(self) -> str | None:
        return str(self.revision_id)

    @override
    @classmethod
    def operation_type(cls) -> ActionOperationType:
        return ActionOperationType.UPDATE


@dataclass
class ActivateRevisionActionResult(BaseActionResult):
    """Result of activating a revision."""

    deployment: ModelDeploymentData
    previous_revision_id: UUID | None
    activated_revision_id: UUID

    @override
    def entity_id(self) -> str | None:
        return str(self.activated_revision_id)



from __future__ import annotations

from dataclasses import dataclass
from typing import override

from ai.backend.common.data.permission.types import EntityType
from ai.backend.manager.actions.action import BaseActionResult
from ai.backend.manager.actions.types import ActionOperationType
from ai.backend.manager.data.kernel.types import KernelInfo
from ai.backend.manager.repositories.base import BatchQuerier
from ai.backend.manager.services.session.base import SessionAction


@dataclass
class SearchKernelsAction(SessionAction):
    querier: BatchQuerier

    @override
    @classmethod
    def entity_type(cls) -> EntityType:
        return EntityType.SESSION_KERNEL

    @override
    def entity_id(self) -> str | None:
        return None

    @override
    @classmethod
    def operation_type(cls) -> ActionOperationType:
        return ActionOperationType.SEARCH


@dataclass
class SearchKernelsActionResult(BaseActionResult):
    data: list[KernelInfo]
    total_count: int
    has_next_page: bool
    has_previous_page: bool

    @override
    def entity_id(self) -> str | None:
        return None



from __future__ import annotations

import logging
import uuid
from typing import TYPE_CHECKING

import sqlalchemy as sa
from sqlalchemy.orm import Mapped, foreign, mapped_column, relationship

from ai.backend.logging import BraceStyleAdapter
from ai.backend.manager.data.storage_namespace.types import StorageNamespaceData
from ai.backend.manager.models.base import (
    GUID,
    Base,
)

if TYPE_CHECKING:
    from ai.backend.manager.models.object_storage import ObjectStorageRow

log = BraceStyleAdapter(logging.getLogger(__spec__.name))

__all__ = ("StorageNamespaceRow",)


def _get_storage_namespace_join_cond() -> sa.ColumnElement[bool]:
    from ai.backend.manager.models.object_storage import ObjectStorageRow

    return foreign(StorageNamespaceRow.storage_id) == ObjectStorageRow.id


class StorageNamespaceRow(Base):  # type: ignore[misc]
    __tablename__ = "storage_namespace"
    __table_args__ = (
        # constraint
        sa.UniqueConstraint("storage_id", "namespace", name="uq_storage_id_namespace"),
    )

    id: Mapped[uuid.UUID] = mapped_column(
        "id", GUID, primary_key=True, server_default=sa.text("uuid_generate_v4()")
    )
    storage_id: Mapped[uuid.UUID] = mapped_column(
        "storage_id",
        GUID,
        nullable=False,
    )
    namespace: Mapped[str] = mapped_column("namespace", sa.String, nullable=False)

    object_storage_row: Mapped[ObjectStorageRow] = relationship(
        "ObjectStorageRow",
        back_populates="namespace_rows",
        primaryjoin=_get_storage_namespace_join_cond,
    )

    def to_dataclass(self) -> StorageNamespaceData:
        return StorageNamespaceData(
            id=self.id,
            storage_id=self.storage_id,
            namespace=self.namespace,
        )



from __future__ import annotations

from dataclasses import dataclass, field
from typing import Any, override

from ai.backend.common.types import BinarySize, ResourceSlot
from ai.backend.manager.models.resource_preset import ResourcePresetRow
from ai.backend.manager.repositories.base.updater import UpdaterSpec
from ai.backend.manager.types import OptionalState, TriState


@dataclass
class ResourcePresetUpdaterSpec(UpdaterSpec[ResourcePresetRow]):
    """UpdaterSpec for resource preset updates."""

    resource_slots: OptionalState[ResourceSlot] = field(
        default_factory=OptionalState[ResourceSlot].nop
    )
    name: OptionalState[str] = field(default_factory=OptionalState[str].nop)
    shared_memory: TriState[BinarySize] = field(default_factory=TriState[BinarySize].nop)
    scaling_group_name: TriState[str] = field(default_factory=TriState[str].nop)

    @property
    @override
    def row_class(self) -> type[ResourcePresetRow]:
        return ResourcePresetRow

    @override
    def build_values(self) -> dict[str, Any]:
        to_update: dict[str, Any] = {}
        self.resource_slots.update_dict(to_update, "resource_slots")
        self.name.update_dict(to_update, "name")
        self.shared_memory.update_dict(to_update, "shared_memory")
        self.scaling_group_name.update_dict(to_update, "scaling_group_name")
        return to_update



"""add_mounts_info_in_kernel_db

Revision ID: 2a82340fa30e
Revises: c1409ad0e8da
Create Date: 2019-08-01 15:59:41.807766

"""

import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "2a82340fa30e"
down_revision = "c1409ad0e8da"
branch_labels = None
depends_on = None


def upgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.add_column("kernels", sa.Column("mounts", sa.ARRAY(sa.String()), nullable=True))
    # ### end Alembic commands ###


def downgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_column("kernels", "mounts")
    # ### end Alembic commands ###



import sqlalchemy as sa

from ai.backend.common.exception import BackendAIError
from ai.backend.common.metrics.metric import DomainType, LayerType
from ai.backend.common.resilience.policies.metrics import MetricArgs, MetricPolicy
from ai.backend.common.resilience.policies.retry import BackoffStrategy, RetryArgs, RetryPolicy
from ai.backend.common.resilience.resilience import Resilience
from ai.backend.manager.data.resource.types import ProjectResourcePolicyData
from ai.backend.manager.errors.common import ObjectNotFound
from ai.backend.manager.models.resource_policy import ProjectResourcePolicyRow
from ai.backend.manager.models.utils import ExtendedAsyncSAEngine
from ai.backend.manager.repositories.base.creator import Creator, execute_creator
from ai.backend.manager.repositories.base.updater import Updater, execute_updater

project_resource_policy_repository_resilience = Resilience(
    policies=[
        MetricPolicy(
            MetricArgs(
                domain=DomainType.REPOSITORY, layer=LayerType.PROJECT_RESOURCE_POLICY_REPOSITORY
            )
        ),
        RetryPolicy(
            RetryArgs(
                max_retries=10,
                retry_delay=0.1,
                backoff_strategy=BackoffStrategy.FIXED,
                non_retryable_exceptions=(BackendAIError,),
            )
        ),
    ]
)


class ProjectResourcePolicyRepository:
    _db: ExtendedAsyncSAEngine

    def __init__(self, db: ExtendedAsyncSAEngine) -> None:
        self._db = db

    @project_resource_policy_repository_resilience.apply()
    async def create(self, creator: Creator[ProjectResourcePolicyRow]) -> ProjectResourcePolicyData:
        async with self._db.begin_session() as db_sess:
            result = await execute_creator(db_sess, creator)
            return result.row.to_dataclass()

    @project_resource_policy_repository_resilience.apply()
    async def get_by_name(self, name: str) -> ProjectResourcePolicyData:
        async with self._db.begin_readonly_session_read_committed() as db_sess:
            query = sa.select(ProjectResourcePolicyRow).where(ProjectResourcePolicyRow.name == name)
            result = await db_sess.execute(query)
            row = result.scalar_one_or_none()
            if row is None:
                raise ObjectNotFound(f"Project resource policy with name {name} not found.")
            return row.to_dataclass()

    @project_resource_policy_repository_resilience.apply()
    async def update(self, updater: Updater[ProjectResourcePolicyRow]) -> ProjectResourcePolicyData:
        async with self._db.begin_session() as db_sess:
            result = await execute_updater(db_sess, updater)
            if result is None:
                raise ObjectNotFound(
                    f"Project resource policy with name {updater.pk_value} not found."
                )
            return result.row.to_dataclass()

    @project_resource_policy_repository_resilience.apply()
    async def delete(self, name: str) -> ProjectResourcePolicyData:
        async with self._db.begin_session() as db_sess:
            query = sa.select(ProjectResourcePolicyRow).where(ProjectResourcePolicyRow.name == name)
            result = await db_sess.execute(query)
            row = result.scalar_one_or_none()
            if row is None:
                raise ObjectNotFound(f"Project resource policy with name {name} not found.")
            await db_sess.delete(row)
            return row.to_dataclass()



"""
REST API handlers for CSV export system.
Provides endpoints for listing reports and report metadata.

Report-specific CSV export handlers are in separate modules:
- users.py: POST /export/users/csv
- sessions.py: POST /export/sessions/csv
- projects.py: POST /export/projects/csv
- keypairs.py: POST /export/keypairs/csv
- audit_logs.py: POST /export/audit-logs/csv
"""

from __future__ import annotations

from collections.abc import Iterable
from http import HTTPStatus

import aiohttp_cors
from aiohttp import web

from ai.backend.common.api_handlers import (
    APIResponse,
    PathParam,
    api_handler,
)
from ai.backend.common.contexts.user import current_user
from ai.backend.common.dto.manager.export import (
    ExportFieldInfo,
    ExportReportInfo,
    GetExportReportResponse,
    ListExportReportsResponse,
)
from ai.backend.manager.api.auth import auth_required_for_method
from ai.backend.manager.api.types import CORSOptions, WebMiddleware
from ai.backend.manager.dto.context import ProcessorsCtx
from ai.backend.manager.dto.export import ExportPathParam
from ai.backend.manager.services.export.actions import (
    GetReportAction,
    ListReportsAction,
)

from .audit_logs import AuditLogExportHandler
from .keypairs import KeypairExportHandler
from .projects import ProjectExportHandler
from .sessions import SessionExportHandler
from .users import UserExportHandler

__all__ = ("create_app",)


class ExportAPIHandler:
    """REST API handler class for CSV export metadata operations.

    Handles listing and getting report metadata.
    Report-specific CSV export is handled by dedicated handlers.
    """

    @auth_required_for_method
    @api_handler
    async def list_reports(
        self,
        processors_ctx: ProcessorsCtx,
    ) -> APIResponse:
        """List available export reports."""
        me = current_user()
        if me is None or not me.is_superadmin:
            raise web.HTTPForbidden(reason="Only superadmin can access export reports.")

        # Call service through processors
        action_result = await processors_ctx.processors.export.list_reports.wait_for_complete(
            ListReportsAction()
        )

        # Convert to response DTO
        reports = [
            ExportReportInfo(
                report_key=r.report_key,
                name=r.name,
                description=r.description,
                fields=[
                    ExportFieldInfo(
                        key=f.key,
                        name=f.name,
                        description=f.description,
                        field_type=f.field_type.value,
                    )
                    for f in r.fields
                ],
            )
            for r in action_result.reports
        ]

        resp = ListExportReportsResponse(reports=reports)
        return APIResponse.build(status_code=HTTPStatus.OK, response_model=resp)

    @auth_required_for_method
    @api_handler
    async def get_report(
        self,
        path: PathParam[ExportPathParam],
        processors_ctx: ProcessorsCtx,
    ) -> APIResponse:
        """Get a specific export report by key."""
        me = current_user()
        if me is None or not me.is_superadmin:
            raise web.HTTPForbidden(reason="Only superadmin can access export reports.")

        # Call service through processors
        action_result = await processors_ctx.processors.export.get_report.wait_for_complete(
            GetReportAction(report_key=path.parsed.report_key)
        )

        # Convert to response DTO
        report = action_result.report
        resp = GetExportReportResponse(
            report=ExportReportInfo(
                report_key=report.report_key,
                name=report.name,
                description=report.description,
                fields=[
                    ExportFieldInfo(
                        key=f.key,
                        name=f.name,
                        description=f.description,
                        field_type=f.field_type.value,
                    )
                    for f in report.fields
                ],
            )
        )
        return APIResponse.build(status_code=HTTPStatus.OK, response_model=resp)


def create_app(
    default_cors_options: CORSOptions,
) -> tuple[web.Application, Iterable[WebMiddleware]]:
    """Create aiohttp application for export API endpoints."""
    app = web.Application()
    app["api_versions"] = (4, 5)
    app["prefix"] = "export"

    cors = aiohttp_cors.setup(app, defaults=default_cors_options)

    # Common handlers for report metadata
    handler = ExportAPIHandler()
    cors.add(app.router.add_route("GET", "/reports", handler.list_reports))
    cors.add(app.router.add_route("GET", "/reports/{report_key}", handler.get_report))

    # Report-specific CSV export handlers
    user_handler = UserExportHandler()
    cors.add(app.router.add_route("POST", "/users/csv", user_handler.export_csv))

    session_handler = SessionExportHandler()
    cors.add(app.router.add_route("POST", "/sessions/csv", session_handler.export_csv))

    project_handler = ProjectExportHandler()
    cors.add(app.router.add_route("POST", "/projects/csv", project_handler.export_csv))

    keypair_handler = KeypairExportHandler()
    cors.add(app.router.add_route("POST", "/keypairs/csv", keypair_handler.export_csv))

    audit_log_handler = AuditLogExportHandler()
    cors.add(app.router.add_route("POST", "/audit-logs/csv", audit_log_handler.export_csv))

    return app, []



from dataclasses import dataclass
from typing import override

from ai.backend.manager.actions.action import BaseActionResult
from ai.backend.manager.actions.types import ActionOperationType
from ai.backend.manager.data.container_registry.types import ContainerRegistryData
from ai.backend.manager.models.container_registry import ContainerRegistryRow
from ai.backend.manager.repositories.base.creator import Creator
from ai.backend.manager.services.container_registry.actions.base import ContainerRegistryAction


@dataclass
class CreateContainerRegistryAction(ContainerRegistryAction):
    creator: Creator[ContainerRegistryRow]

    @override
    def entity_id(self) -> str | None:
        return None

    @override
    @classmethod
    def operation_type(cls) -> ActionOperationType:
        return ActionOperationType.CREATE


@dataclass
class CreateContainerRegistryActionResult(BaseActionResult):
    data: ContainerRegistryData

    @override
    def entity_id(self) -> str | None:
        return str(self.data.id)



"""
REST API handlers for scheduling history.
Provides search endpoints for session, deployment, and route scheduling history.
"""

from __future__ import annotations

from collections.abc import Iterable
from http import HTTPStatus

import aiohttp_cors
from aiohttp import web

from ai.backend.common.api_handlers import APIResponse, BodyParam, api_handler
from ai.backend.common.contexts.user import current_user
from ai.backend.common.dto.manager.scheduling_history import (
    ListDeploymentHistoryResponse,
    ListRouteHistoryResponse,
    ListSessionHistoryResponse,
    PaginationInfo,
    SearchDeploymentHistoryRequest,
    SearchRouteHistoryRequest,
    SearchSessionHistoryRequest,
)
from ai.backend.manager.api.auth import auth_required_for_method
from ai.backend.manager.api.types import CORSOptions, WebMiddleware
from ai.backend.manager.dto.context import ProcessorsCtx
from ai.backend.manager.services.scheduling_history.actions import (
    SearchDeploymentHistoryAction,
    SearchRouteHistoryAction,
    SearchSessionHistoryAction,
)

from .adapter import SchedulingHistoryAdapter

__all__ = ("create_app",)


class SchedulingHistoryAPIHandler:
    """REST API handler class for scheduling history operations."""

    def __init__(self) -> None:
        self._adapter = SchedulingHistoryAdapter()

    @auth_required_for_method
    @api_handler
    async def search_session_history(
        self,
        body: BodyParam[SearchSessionHistoryRequest],
        processors_ctx: ProcessorsCtx,
    ) -> APIResponse:
        """Search session scheduling history."""
        processors = processors_ctx.processors
        me = current_user()
        if me is None or not me.is_superadmin:
            raise web.HTTPForbidden(reason="Only superadmin can search scheduling history.")

        querier = self._adapter.build_session_history_querier(body.parsed)

        action_result = (
            await processors.scheduling_history.search_session_history.wait_for_complete(
                SearchSessionHistoryAction(querier=querier)
            )
        )

        resp = ListSessionHistoryResponse(
            items=[
                self._adapter.convert_session_history_to_dto(h) for h in action_result.histories
            ],
            pagination=PaginationInfo(
                total=action_result.total_count,
                offset=body.parsed.offset,
                limit=body.parsed.limit,
            ),
        )
        return APIResponse.build(status_code=HTTPStatus.OK, response_model=resp)

    @auth_required_for_method
    @api_handler
    async def search_deployment_history(
        self,
        body: BodyParam[SearchDeploymentHistoryRequest],
        processors_ctx: ProcessorsCtx,
    ) -> APIResponse:
        """Search deployment history."""
        processors = processors_ctx.processors
        me = current_user()
        if me is None or not me.is_superadmin:
            raise web.HTTPForbidden(reason="Only superadmin can search scheduling history.")

        querier = self._adapter.build_deployment_history_querier(body.parsed)

        action_result = (
            await processors.scheduling_history.search_deployment_history.wait_for_complete(
                SearchDeploymentHistoryAction(querier=querier)
            )
        )

        resp = ListDeploymentHistoryResponse(
            items=[
                self._adapter.convert_deployment_history_to_dto(h) for h in action_result.histories
            ],
            pagination=PaginationInfo(
                total=action_result.total_count,
                offset=body.parsed.offset,
                limit=body.parsed.limit,
            ),
        )
        return APIResponse.build(status_code=HTTPStatus.OK, response_model=resp)

    @auth_required_for_method
    @api_handler
    async def search_route_history(
        self,
        body: BodyParam[SearchRouteHistoryRequest],
        processors_ctx: ProcessorsCtx,
    ) -> APIResponse:
        """Search route history."""
        processors = processors_ctx.processors
        me = current_user()
        if me is None or not me.is_superadmin:
            raise web.HTTPForbidden(reason="Only superadmin can search scheduling history.")

        querier = self._adapter.build_route_history_querier(body.parsed)

        action_result = await processors.scheduling_history.search_route_history.wait_for_complete(
            SearchRouteHistoryAction(querier=querier)
        )

        resp = ListRouteHistoryResponse(
            items=[self._adapter.convert_route_history_to_dto(h) for h in action_result.histories],
            pagination=PaginationInfo(
                total=action_result.total_count,
                offset=body.parsed.offset,
                limit=body.parsed.limit,
            ),
        )
        return APIResponse.build(status_code=HTTPStatus.OK, response_model=resp)


def create_app(
    default_cors_options: CORSOptions,
) -> tuple[web.Application, Iterable[WebMiddleware]]:
    """Create aiohttp application for scheduling history API endpoints."""
    app = web.Application()
    app["api_versions"] = (4, 5)
    app["prefix"] = "scheduling-history"

    cors = aiohttp_cors.setup(app, defaults=default_cors_options)
    api_handler = SchedulingHistoryAPIHandler()

    # Session history routes
    cors.add(app.router.add_route("POST", "/sessions/search", api_handler.search_session_history))

    # Deployment history routes
    cors.add(
        app.router.add_route("POST", "/deployments/search", api_handler.search_deployment_history)
    )

    # Route history routes
    cors.add(app.router.add_route("POST", "/routes/search", api_handler.search_route_history))

    return app, []



from collections.abc import Mapping
from dataclasses import dataclass
from typing import Any, override

from ai.backend.common.dto.agent.response import CodeCompletionResp
from ai.backend.common.types import AccessKey
from ai.backend.manager.actions.action import BaseActionResult
from ai.backend.manager.actions.types import ActionOperationType
from ai.backend.manager.data.session.types import SessionData
from ai.backend.manager.services.session.base import SessionAction


# TODO: Rename this?
@dataclass
class CompleteAction(SessionAction):
    session_name: str
    owner_access_key: AccessKey
    code: str
    # TODO: Add type
    options: Mapping[str, Any] | None

    @override
    def entity_id(self) -> str | None:
        return None

    @override
    @classmethod
    def operation_type(cls) -> ActionOperationType:
        return ActionOperationType.UPDATE


@dataclass
class CompleteActionResult(BaseActionResult):
    session_data: SessionData

    result: CodeCompletionResp

    @override
    def entity_id(self) -> str | None:
        return str(self.session_data.id)



from __future__ import annotations

from typing import TYPE_CHECKING

import sqlalchemy as sa

from ai.backend.common.exception import BackendAIError
from ai.backend.common.metrics.metric import DomainType, LayerType
from ai.backend.common.resilience.policies.metrics import MetricArgs, MetricPolicy
from ai.backend.common.resilience.policies.retry import BackoffStrategy, RetryArgs, RetryPolicy
from ai.backend.common.resilience.resilience import Resilience
from ai.backend.manager.data.audit_log.types import AuditLogData, AuditLogListResult
from ai.backend.manager.models.audit_log import AuditLogRow
from ai.backend.manager.repositories.base import (
    BatchQuerier,
    Creator,
    execute_batch_querier,
    execute_creator,
)

if TYPE_CHECKING:
    from ai.backend.manager.models.utils import ExtendedAsyncSAEngine

__all__ = ("AuditLogDBSource",)

audit_log_db_source_resilience = Resilience(
    policies=[
        MetricPolicy(MetricArgs(domain=DomainType.DB_SOURCE, layer=LayerType.AUDIT_LOG_DB_SOURCE)),
        RetryPolicy(
            RetryArgs(
                max_retries=5,
                retry_delay=0.1,
                backoff_strategy=BackoffStrategy.FIXED,
                non_retryable_exceptions=(BackendAIError,),
            )
        ),
    ]
)


class AuditLogDBSource:
    _db: ExtendedAsyncSAEngine

    def __init__(self, db: ExtendedAsyncSAEngine) -> None:
        self._db = db

    @audit_log_db_source_resilience.apply()
    async def create(self, creator: Creator[AuditLogRow]) -> AuditLogData:
        async with self._db.begin_session() as db_sess:
            result = await execute_creator(db_sess, creator)
            return result.row.to_dataclass()

    @audit_log_db_source_resilience.apply()
    async def search(
        self,
        querier: BatchQuerier,
    ) -> AuditLogListResult:
        """Search audit logs with querier pattern.

        Args:
            querier: BatchQuerier for filtering, ordering, and pagination

        Returns:
            AuditLogListResult with items, total count, and pagination info
        """
        async with self._db.begin_readonly_session() as db_sess:
            query = sa.select(AuditLogRow)

            result = await execute_batch_querier(
                db_sess,
                query,
                querier,
            )

            items = [row.AuditLogRow.to_dataclass() for row in result.rows]

            return AuditLogListResult(
                items=items,
                total_count=result.total_count,
                has_next_page=result.has_next_page,
                has_previous_page=result.has_previous_page,
            )



from __future__ import annotations

from ai.backend.common.configs.migration.types import MappingRule

MANAGER_MAPPING_RULES: list[MappingRule] = [
    MappingRule(
        source="manager.service-addr",
        role="api",
        scope="cluster",
        protocol="http",
    ),
]



"""CreatorSpec for deployment auto-scaling policy creation."""

from __future__ import annotations

import uuid
from dataclasses import dataclass
from decimal import Decimal
from typing import override

from ai.backend.common.types import (
    AutoScalingMetricComparator,
    AutoScalingMetricSource,
)
from ai.backend.manager.models.deployment_auto_scaling_policy import (
    DeploymentAutoScalingPolicyRow,
)
from ai.backend.manager.repositories.base import CreatorSpec


@dataclass
class DeploymentAutoScalingPolicyCreatorSpec(CreatorSpec[DeploymentAutoScalingPolicyRow]):
    """CreatorSpec for deployment auto-scaling policy creation.

    Each endpoint can have at most one auto-scaling policy (1:1 relationship).
    The policy supports dual thresholds for hysteresis-based scaling.
    """

    endpoint_id: uuid.UUID
    min_replicas: int
    max_replicas: int
    metric_source: AutoScalingMetricSource | None
    metric_name: str | None
    comparator: AutoScalingMetricComparator | None
    scale_up_threshold: Decimal | None
    scale_down_threshold: Decimal | None
    scale_up_step_size: int
    scale_down_step_size: int
    cooldown_seconds: int

    @override
    def build_row(self) -> DeploymentAutoScalingPolicyRow:
        return DeploymentAutoScalingPolicyRow(
            endpoint=self.endpoint_id,
            min_replicas=self.min_replicas,
            max_replicas=self.max_replicas,
            metric_source=self.metric_source,
            metric_name=self.metric_name,
            comparator=self.comparator,
            scale_up_threshold=self.scale_up_threshold,
            scale_down_threshold=self.scale_down_threshold,
            scale_up_step_size=self.scale_up_step_size,
            scale_down_step_size=self.scale_down_step_size,
            cooldown_seconds=self.cooldown_seconds,
        )



"""CreatorSpec and BatchUpdaterSpec for route creation and updates."""

from __future__ import annotations

import uuid
from dataclasses import dataclass
from typing import Any, override

from ai.backend.manager.data.deployment.types import RouteStatus, RouteTrafficStatus
from ai.backend.manager.models.routing import RoutingRow
from ai.backend.manager.repositories.base import CreatorSpec
from ai.backend.manager.repositories.base.updater import BatchUpdaterSpec


@dataclass
class RouteCreatorSpec(CreatorSpec[RoutingRow]):
    """CreatorSpec for route creation.

    Routes are created for deployment endpoints to serve traffic.
    Each route can be associated with a specific revision.
    """

    endpoint_id: uuid.UUID
    session_owner_id: uuid.UUID
    domain: str
    project_id: uuid.UUID
    traffic_ratio: float = 1.0
    revision_id: uuid.UUID | None = None
    traffic_status: RouteTrafficStatus = RouteTrafficStatus.ACTIVE

    @override
    def build_row(self) -> RoutingRow:
        return RoutingRow(
            id=uuid.uuid4(),
            endpoint=self.endpoint_id,
            session=None,
            session_owner=self.session_owner_id,
            domain=self.domain,
            project=self.project_id,
            status=RouteStatus.PROVISIONING,
            traffic_ratio=self.traffic_ratio,
            revision=self.revision_id,
            traffic_status=self.traffic_status,
        )


@dataclass
class RouteBatchUpdaterSpec(BatchUpdaterSpec[RoutingRow]):
    """BatchUpdaterSpec for batch updating routes.

    Accepts optional fields and only updates fields that are specified.
    This allows flexible partial updates for various route operations.
    """

    status: RouteStatus | None = None
    traffic_ratio: float | None = None
    traffic_status: RouteTrafficStatus | None = None

    @property
    @override
    def row_class(self) -> type[RoutingRow]:
        return RoutingRow

    @override
    def build_values(self) -> dict[str, Any]:
        values: dict[str, Any] = {}
        if self.status is not None:
            values["status"] = self.status
        if self.traffic_ratio is not None:
            values["traffic_ratio"] = self.traffic_ratio
        if self.traffic_status is not None:
            values["traffic_status"] = self.traffic_status
        return values



from __future__ import annotations

import uuid
from collections.abc import Sequence

from ai.backend.manager.data.error_log.types import ErrorLogData
from ai.backend.manager.repositories.base import BatchQuerier, OffsetPagination
from ai.backend.manager.repositories.error_log import ErrorLogConditions
from ai.backend.manager.services.error_log.actions.search import SearchErrorLogsAction
from ai.backend.manager.services.error_log.processors import ErrorLogProcessors


async def load_error_logs_by_ids(
    processor: ErrorLogProcessors,
    error_log_ids: Sequence[uuid.UUID],
) -> list[ErrorLogData | None]:
    """Batch load error logs by their IDs.

    Args:
        processor: The error log processor.
        error_log_ids: Sequence of error log UUIDs to load.

    Returns:
        List of ErrorLogData (or None if not found) in the same order as error_log_ids.
    """
    if not error_log_ids:
        return []

    querier = BatchQuerier(
        pagination=OffsetPagination(limit=len(error_log_ids)),
        conditions=[ErrorLogConditions.by_ids(error_log_ids)],
    )

    action_result = await processor.search.wait_for_complete(SearchErrorLogsAction(querier=querier))

    error_log_map = {error_log.id: error_log for error_log in action_result.data}
    return [error_log_map.get(error_log_id) for error_log_id in error_log_ids]



"""Main deployment repository implementation."""

import logging
import uuid
from collections import defaultdict
from collections.abc import Mapping, Sequence
from dataclasses import dataclass, field
from datetime import UTC, datetime, timedelta
from decimal import Decimal, DecimalException
from typing import Any, cast

import tomli
from pydantic import HttpUrl
from ruamel.yaml import YAML

from ai.backend.common.clients.valkey_client.valkey_live.client import ValkeyLiveClient
from ai.backend.common.clients.valkey_client.valkey_schedule.client import ValkeyScheduleClient
from ai.backend.common.clients.valkey_client.valkey_stat.client import ValkeyStatClient
from ai.backend.common.data.endpoint.types import EndpointLifecycle
from ai.backend.common.exception import BackendAIError, InvalidAPIParameters
from ai.backend.common.metrics.metric import DomainType, LayerType
from ai.backend.common.resilience.policies.metrics import MetricArgs, MetricPolicy
from ai.backend.common.resilience.policies.retry import BackoffStrategy, RetryArgs, RetryPolicy
from ai.backend.common.resilience.resilience import Resilience
from ai.backend.common.types import (
    AutoScalingMetricComparator,
    AutoScalingMetricSource,
    KernelId,
    SessionId,
)
from ai.backend.logging.utils import BraceStyleAdapter
from ai.backend.manager.api.gql_legacy.statistics import EndpointStatistics, KernelStatistics
from ai.backend.manager.data.deployment.creator import DeploymentPolicyConfig
from ai.backend.manager.data.deployment.scale import (
    AutoScalingRule,
    AutoScalingRuleCreator,
    ModelDeploymentAutoScalingRuleCreator,
)
from ai.backend.manager.data.deployment.scale_modifier import (
    AutoScalingRuleModifier,
    ModelDeploymentAutoScalingRuleModifier,
)
from ai.backend.manager.data.deployment.types import (
    AccessTokenSearchResult,
    AutoScalingRuleSearchResult,
    DefinitionFiles,
    DeploymentInfo,
    DeploymentInfoSearchResult,
    DeploymentInfoWithAutoScalingRules,
    ModelDeploymentAutoScalingRuleData,
    ModelRevisionData,
    RevisionSearchResult,
    RouteInfo,
    RouteSearchResult,
    RouteStatus,
    ScalingGroupCleanupConfig,
)
from ai.backend.manager.data.image.types import ImageIdentifier
from ai.backend.manager.data.resource.types import ScalingGroupProxyTarget
from ai.backend.manager.data.session.types import SessionStatus
from ai.backend.manager.errors.deployment import DefinitionFileNotFound
from ai.backend.manager.errors.service import EndpointNotFound
from ai.backend.manager.models.deployment_auto_scaling_policy import (
    DeploymentAutoScalingPolicyData,
    DeploymentAutoScalingPolicyRow,
)
from ai.backend.manager.models.deployment_policy import (
    DeploymentPolicyData,
    DeploymentPolicyRow,
)
from ai.backend.manager.models.deployment_revision import DeploymentRevisionRow
from ai.backend.manager.models.endpoint import EndpointRow, EndpointTokenRow
from ai.backend.manager.models.routing import RoutingRow
from ai.backend.manager.models.scheduling_history import (
    DeploymentHistoryRow,
    RouteHistoryRow,
)
from ai.backend.manager.models.storage import StorageSessionManager
from ai.backend.manager.models.utils import ExtendedAsyncSAEngine
from ai.backend.manager.models.vfolder import VFolderOwnershipType
from ai.backend.manager.repositories.base import BatchQuerier, Creator
from ai.backend.manager.repositories.base.creator import BulkCreator
from ai.backend.manager.repositories.base.purger import Purger, PurgerResult
from ai.backend.manager.repositories.base.rbac.entity_creator import RBACEntityCreator
from ai.backend.manager.repositories.base.updater import BatchUpdater, Updater
from ai.backend.manager.repositories.scheduler.types.session_creation import DeploymentContext

from .db_source import DeploymentDBSource
from .storage_source import DeploymentStorageSource
from .types import RouteData, RouteServiceDiscoveryInfo

log = BraceStyleAdapter(logging.getLogger(__name__))


@dataclass
class AutoScalingMetricsData:
    """Container for all metrics data needed for auto-scaling calculations."""

    kernel_statistics: dict[KernelId, Mapping[str, Any] | None] = field(default_factory=dict)
    endpoint_statistics: dict[uuid.UUID, Mapping[str, Any] | None] = field(default_factory=dict)
    routes_by_endpoint: Mapping[uuid.UUID, list[RouteInfo]] = field(default_factory=dict)
    kernels_by_session: dict[SessionId, list[KernelId]] = field(default_factory=dict)


deployment_repository_resilience = Resilience(
    policies=[
        MetricPolicy(
            MetricArgs(domain=DomainType.REPOSITORY, layer=LayerType.DEPLOYMENT_REPOSITORY)
        ),
        RetryPolicy(
            RetryArgs(
                max_retries=3,
                retry_delay=0.1,
                backoff_strategy=BackoffStrategy.FIXED,
                non_retryable_exceptions=(BackendAIError,),
            )
        ),
    ]
)


class DeploymentRepository:
    """Repository for deployment-related operations."""

    _db_source: DeploymentDBSource
    _storage_source: DeploymentStorageSource
    _valkey_stat: ValkeyStatClient
    _valkey_live: ValkeyLiveClient
    _valkey_schedule: ValkeyScheduleClient

    def __init__(
        self,
        db: ExtendedAsyncSAEngine,
        storage_manager: StorageSessionManager,
        valkey_stat: ValkeyStatClient,
        valkey_live: ValkeyLiveClient,
        valkey_schedule: ValkeyScheduleClient,
    ) -> None:
        self._db_source = DeploymentDBSource(db, storage_manager)
        self._storage_source = DeploymentStorageSource(storage_manager)
        self._valkey_stat = valkey_stat
        self._valkey_live = valkey_live
        self._valkey_schedule = valkey_schedule

    # Endpoint operations

    @deployment_repository_resilience.apply()
    async def create_endpoint(
        self,
        creator: RBACEntityCreator[EndpointRow],
        policy_config: DeploymentPolicyConfig | None = None,
    ) -> DeploymentInfo:
        """Create a new endpoint and return DeploymentInfo.

        Args:
            creator: Creator containing DeploymentCreatorSpec with resolved image_id
            policy_config: Optional deployment policy configuration

        Returns:
            DeploymentInfo for the created endpoint
        """
        return await self._db_source.create_endpoint(creator, policy_config)

    @deployment_repository_resilience.apply()
    async def create_endpoint_legacy(
        self,
        creator: RBACEntityCreator[EndpointRow],
    ) -> DeploymentInfo:
        """Create a new endpoint using legacy DeploymentCreator.

        This is for backward compatibility with legacy deployment creation flow.

        Args:
            creator: RBACEntityCreator with LegacyEndpointCreatorSpec.
                The spec MUST be an instance of LegacyEndpointCreatorSpec.

        Returns:
            DeploymentInfo for the created endpoint
        """
        return await self._db_source.create_endpoint_legacy(creator)

    @deployment_repository_resilience.apply()
    async def get_image_id(self, image: ImageIdentifier) -> uuid.UUID:
        """Get image ID from ImageIdentifier."""
        return await self._db_source.get_image_id(image)

    @deployment_repository_resilience.apply()
    async def get_modified_endpoint(
        self,
        endpoint_id: uuid.UUID,
        updater: Updater[EndpointRow],
    ) -> DeploymentInfo:
        """Get modified endpoint without applying changes.

        Args:
            endpoint_id: ID of the endpoint to modify
            updater: Updater containing spec with partial updates

        Returns:
            DeploymentInfo: Modified deployment information

        Raises:
            EndpointNotFound: If the endpoint does not exist
        """
        return await self._db_source.get_modified_endpoint(endpoint_id, updater)

    @deployment_repository_resilience.apply()
    async def update_endpoint_with_spec(
        self,
        updater: Updater[EndpointRow],
    ) -> DeploymentInfo:
        """Update endpoint using an Updater.

        Args:
            updater: Updater containing spec and endpoint_id

        Returns:
            DeploymentInfo: Updated deployment information

        Raises:
            NoUpdatesToApply: If there are no updates to apply
            EndpointNotFound: If the endpoint does not exist
        """
        return await self._db_source.update_endpoint_with_spec(updater)

    @deployment_repository_resilience.apply()
    async def update_endpoint_lifecycle_bulk(
        self,
        endpoint_ids: list[uuid.UUID],
        prevoius_status: list[EndpointLifecycle],
        new_status: EndpointLifecycle,
    ) -> None:
        """Update lifecycle status for multiple endpoints."""
        await self._db_source.update_endpoint_lifecycle_bulk(
            endpoint_ids, prevoius_status, new_status
        )

    @deployment_repository_resilience.apply()
    async def update_endpoint_lifecycle_bulk_with_history(
        self,
        batch_updaters: Sequence[BatchUpdater[EndpointRow]],
        bulk_creator: BulkCreator[DeploymentHistoryRow],
    ) -> int:
        """Update lifecycle status and record history in same transaction.

        All batch updates and history creations are executed atomically
        in a single transaction.

        Args:
            batch_updaters: Sequence of BatchUpdaters for status updates
            bulk_creator: BulkCreator containing all history records

        Returns:
            Total number of rows updated
        """
        return await self._db_source.update_endpoint_lifecycle_bulk_with_history(
            batch_updaters, bulk_creator
        )

    @deployment_repository_resilience.apply()
    async def get_endpoints_by_ids(
        self,
        endpoint_ids: set[uuid.UUID],
    ) -> list[DeploymentInfo]:
        """Get endpoints by their IDs."""
        return await self._db_source.get_endpoints_by_ids(endpoint_ids)

    @deployment_repository_resilience.apply()
    async def get_scaling_group_cleanup_configs(
        self, scaling_group_names: Sequence[str]
    ) -> Mapping[str, ScalingGroupCleanupConfig]:
        """
        Get route cleanup target statuses configuration for scaling groups.

        Args:
            scaling_group_names: List of scaling group names to query

        Returns:
            Mapping of scaling group name to ScalingGroupCleanupConfig
        """
        return await self._db_source.get_scaling_group_cleanup_configs(scaling_group_names)

    @deployment_repository_resilience.apply()
    async def get_endpoints_by_statuses(
        self,
        statuses: list[EndpointLifecycle],
    ) -> list[DeploymentInfo]:
        """Get endpoints by lifecycle statuses."""
        return await self._db_source.get_endpoints_by_statuses(statuses)

    @deployment_repository_resilience.apply()
    async def get_endpoint_info(
        self,
        endpoint_id: uuid.UUID,
    ) -> DeploymentInfo:
        """Get endpoint information.

        Raises:
            EndpointNotFound: If the endpoint does not exist
        """
        return await self._db_source.get_endpoint(endpoint_id)

    @deployment_repository_resilience.apply()
    async def destroy_endpoint(
        self,
        endpoint_id: uuid.UUID,
    ) -> bool:
        """Destroy an endpoint and all its routes."""
        return await self._db_source.update_endpoint_lifecycle(
            endpoint_id, EndpointLifecycle.DESTROYING
        )

    @deployment_repository_resilience.apply()
    async def delete_endpoint(
        self,
        endpoint_id: uuid.UUID,
    ) -> bool:
        """Delete an endpoint and all its routes."""
        return await self._db_source.delete_endpoint_with_routes(endpoint_id)

    @deployment_repository_resilience.apply()
    async def get_service_endpoint(
        self,
        endpoint_id: uuid.UUID,
    ) -> HttpUrl | None:
        """Get service endpoint URL."""
        try:
            endpoint = await self._db_source.get_endpoint(endpoint_id)
            if not endpoint.network.url:
                return None
            return HttpUrl(endpoint.network.url)
        except EndpointNotFound:
            return None

    # Route operations

    @deployment_repository_resilience.apply()
    async def create_autoscaling_rule(
        self,
        endpoint_id: uuid.UUID,
        creator: AutoScalingRuleCreator,
    ) -> AutoScalingRule:
        """Create a new autoscaling rule for an endpoint."""
        return await self._db_source.create_autoscaling_rule(endpoint_id, creator)

    @deployment_repository_resilience.apply()
    async def list_autoscaling_rules(
        self,
        endpoint_id: uuid.UUID,
    ) -> list[AutoScalingRule]:
        """List all autoscaling rules for an endpoint."""
        return await self._db_source.list_autoscaling_rules(endpoint_id)

    @deployment_repository_resilience.apply()
    async def update_autoscaling_rule(
        self,
        rule_id: uuid.UUID,
        modifier: AutoScalingRuleModifier,
    ) -> AutoScalingRule:
        """Update an existing autoscaling rule."""
        return await self._db_source.update_autoscaling_rule(rule_id, modifier)

    @deployment_repository_resilience.apply()
    async def delete_autoscaling_rule(
        self,
        rule_id: uuid.UUID,
    ) -> bool:
        """Delete an autoscaling rule."""
        return await self._db_source.delete_autoscaling_rule(rule_id)

    # Model Deployment Auto-scaling Rule operations (new types)

    @deployment_repository_resilience.apply()
    async def create_model_deployment_autoscaling_rule(
        self,
        creator: ModelDeploymentAutoScalingRuleCreator,
    ) -> ModelDeploymentAutoScalingRuleData:
        """Create a new autoscaling rule using ModelDeployment types."""
        return await self._db_source.create_model_deployment_autoscaling_rule(creator)

    @deployment_repository_resilience.apply()
    async def update_model_deployment_autoscaling_rule(
        self,
        rule_id: uuid.UUID,
        modifier: ModelDeploymentAutoScalingRuleModifier,
    ) -> ModelDeploymentAutoScalingRuleData:
        """Update an autoscaling rule using ModelDeployment types."""
        return await self._db_source.update_model_deployment_autoscaling_rule(rule_id, modifier)

    @deployment_repository_resilience.apply()
    async def list_model_deployment_autoscaling_rules(
        self,
        endpoint_id: uuid.UUID,
    ) -> list[ModelDeploymentAutoScalingRuleData]:
        """List all autoscaling rules for an endpoint using ModelDeployment types."""
        return await self._db_source.list_model_deployment_autoscaling_rules(endpoint_id)

    @deployment_repository_resilience.apply()
    async def get_model_deployment_autoscaling_rule(
        self,
        rule_id: uuid.UUID,
    ) -> ModelDeploymentAutoScalingRuleData:
        """Get a single autoscaling rule by ID using ModelDeployment types."""
        return await self._db_source.get_model_deployment_autoscaling_rule(rule_id)

    # Data fetching operations
    @deployment_repository_resilience.apply()
    async def fetch_model_definition(
        self,
        vfolder_id: uuid.UUID,
        model_definition_path: str | None,
    ) -> dict[str, Any]:
        """
        Fetch model definition file from model vfolder.

        Args:
            vfolder_id: ID of the model vfolder
            definition_path: Path to the model definition file
        Returns:
            dict: Parsed model definition content
        """
        vfolder_location = await self._db_source.get_vfolder_by_id(vfolder_id)
        if vfolder_location.ownership_type == VFolderOwnershipType.GROUP:
            raise InvalidAPIParameters(
                "Cannot create model service with the project type's vfolder"
            )
        model_definition_candidates = (
            [
                model_definition_path,
            ]
            if model_definition_path
            else [
                "model-definition.yaml",
                "model-definition.yml",
            ]
        )
        model_definition_bytes = await self._storage_source.fetch_definition_file(
            vfolder_location,
            model_definition_candidates,
        )
        yaml = YAML()
        return cast(dict[str, Any], yaml.load(model_definition_bytes))

    @deployment_repository_resilience.apply()
    async def fetch_service_definition(
        self,
        vfolder_id: uuid.UUID,
    ) -> dict[str, Any] | None:
        """
        Fetch service definition file from model vfolder.

        Args:
            vfolder_id: ID of the model vfolder
        Returns:
            dict: Parsed service definition content
        """
        vfolder_location = await self._db_source.get_vfolder_by_id(vfolder_id)
        if vfolder_location.ownership_type == VFolderOwnershipType.GROUP:
            raise InvalidAPIParameters(
                "Cannot create model service with the project type's vfolder"
            )

        # Read service definition from storage
        service_definition_content: dict[str, Any] | None = None
        try:
            service_definition_bytes = await self._storage_source.fetch_definition_file(
                vfolder_location,
                ["service-definition.toml"],
            )
            service_definition_content = tomli.loads(service_definition_bytes.decode("utf-8"))
        except DefinitionFileNotFound:
            # Service definition is optional
            pass

        return service_definition_content

    @deployment_repository_resilience.apply()
    async def fetch_definition_files(
        self,
        vfolder_id: uuid.UUID,
        model_definition_path: str | None,
    ) -> DefinitionFiles:
        """
        Fetch definition files(Both service and model definitions) from model vfolder.

        Args:
            vfolder_id: ID of the model vfolder
            definition_path: Path to the definition file
        Returns:
            DefinitionFiles: Contains service definition and model definition bytes
        """
        model_definition_content: dict[str, Any] = await self.fetch_model_definition(
            vfolder_id, model_definition_path
        )
        service_definition_content: dict[str, Any] | None = await self.fetch_service_definition(
            vfolder_id
        )

        return DefinitionFiles(
            service_definition=service_definition_content,
            model_definition=model_definition_content,
        )

    @deployment_repository_resilience.apply()
    async def get_endpoints_with_autoscaling_rules(
        self,
    ) -> list[DeploymentInfoWithAutoScalingRules]:
        """Get endpoints that have autoscaling rules."""
        return await self._db_source.get_endpoints_with_autoscaling_rules()

    @deployment_repository_resilience.apply()
    async def update_autoscaling_rule_triggered(
        self,
        rule_id: uuid.UUID,
        triggered_at: datetime,
    ) -> bool:
        """Update the last triggered time for an autoscaling rule."""
        return await self._db_source.update_autoscaling_rule_triggered(rule_id, triggered_at)

    @deployment_repository_resilience.apply()
    async def batch_update_desired_replicas(
        self,
        updates: dict[uuid.UUID, int],
    ) -> None:
        """Batch update desired replicas for multiple endpoints."""
        return await self._db_source.batch_update_desired_replicas(updates)

    @deployment_repository_resilience.apply()
    async def fetch_scaling_group_proxy_targets(
        self,
        scaling_group: set[str],
    ) -> Mapping[str, ScalingGroupProxyTarget | None]:
        """Fetch the proxy target URL for a scaling group endpoint."""
        return await self._db_source.fetch_scaling_group_proxy_targets(scaling_group)

    @deployment_repository_resilience.apply()
    async def fetch_auto_scaling_rules_by_endpoint_ids(
        self,
        endpoint_ids: set[uuid.UUID],
    ) -> Mapping[uuid.UUID, list[AutoScalingRule]]:
        """Fetch autoscaling rules for multiple endpoints."""
        return await self._db_source.fetch_auto_scaling_rules_by_endpoint_ids(endpoint_ids)

    @deployment_repository_resilience.apply()
    async def fetch_active_routes_by_endpoint_ids(
        self,
        endpoint_ids: set[uuid.UUID],
    ) -> Mapping[uuid.UUID, list[RouteInfo]]:
        """Fetch routes for multiple endpoints."""
        return await self._db_source.fetch_active_routes_by_endpoint_ids(endpoint_ids)

    @deployment_repository_resilience.apply()
    async def scale_routes(
        self,
        scale_out_creators: Sequence[Creator[RoutingRow]],
        scale_in_updater: BatchUpdater[RoutingRow] | None,
    ) -> None:
        await self._db_source.scale_routes(scale_out_creators, scale_in_updater)

    # Route operations

    @deployment_repository_resilience.apply()
    async def get_routes_by_statuses(
        self,
        statuses: list[RouteStatus],
    ) -> list[RouteData]:
        """Get routes by their statuses.

        Args:
            statuses: List of route statuses to filter by

        Returns:
            List of RouteData objects matching the statuses
        """
        return await self._db_source.get_routes_by_statuses(statuses)

    @deployment_repository_resilience.apply()
    async def update_route_status_bulk(
        self,
        route_ids: set[uuid.UUID],
        previous_statuses: list[RouteStatus],
        new_status: RouteStatus,
    ) -> None:
        """Update status for multiple routes.

        Args:
            route_ids: IDs of routes to update
            previous_statuses: Current statuses to validate against
            new_status: New status to set
        """
        await self._db_source.update_route_status_bulk(route_ids, previous_statuses, new_status)

    @deployment_repository_resilience.apply()
    async def update_route_status_bulk_with_history(
        self,
        batch_updaters: Sequence[BatchUpdater[RoutingRow]],
        bulk_creator: BulkCreator[RouteHistoryRow],
    ) -> int:
        """Update route status and record history in same transaction.

        All batch updates and history creations are executed atomically
        in a single transaction.

        Args:
            batch_updaters: Sequence of BatchUpdaters for status updates
            bulk_creator: BulkCreator containing all history records

        Returns:
            Total number of rows updated
        """
        return await self._db_source.update_route_status_bulk_with_history(
            batch_updaters, bulk_creator
        )

    @deployment_repository_resilience.apply()
    async def mark_terminating_route_status_bulk(
        self,
        route_ids: set[uuid.UUID],
    ) -> None:
        """Update status for multiple routes.

        Args:
            route_ids: IDs of routes to update
            previous_statuses: Current statuses to validate against
            new_status: New status to set
        """
        await self._db_source.mark_terminating_route_status_bulk(route_ids)

    @deployment_repository_resilience.apply()
    async def update_desired_replicas_bulk(
        self,
        replica_updates: Mapping[uuid.UUID, int],
    ) -> None:
        """Update desired replicas for multiple endpoints.

        Args:
            replica_updates: Mapping of endpoint IDs to new desired replica counts
        """
        await self._db_source.update_desired_replicas_bulk(replica_updates)

    @deployment_repository_resilience.apply()
    async def update_endpoint_urls_bulk(
        self,
        url_updates: Mapping[uuid.UUID, str],
    ) -> None:
        """Update endpoint URLs for multiple endpoints.

        Args:
            url_updates: Mapping of endpoint IDs to their registered URLs
        """
        await self._db_source.update_endpoint_urls_bulk(url_updates)

    @deployment_repository_resilience.apply()
    async def update_route_sessions(
        self,
        route_session_ids: Mapping[uuid.UUID, SessionId],
    ) -> None:
        """Update session IDs for multiple routes and initialize their health status.

        Args:
            route_session_ids: Mapping of route IDs to new session IDs
        """
        # Update sessions in database
        await self._db_source.update_route_sessions(route_session_ids)
        route_id_strings = [str(route_id) for route_id in route_session_ids.keys()]
        await self._valkey_schedule.initialize_routes_health_status_batch(route_id_strings)

    @deployment_repository_resilience.apply()
    async def delete_routes_by_route_ids(
        self,
        route_ids: set[uuid.UUID],
    ) -> None:
        """Delete routes by their IDs.

        Args:
            route_ids: List of route IDs to delete
        """
        await self._db_source.delete_routes_by_route_ids(route_ids)

    @deployment_repository_resilience.apply()
    async def fetch_deployment_context(
        self,
        deployment_info: DeploymentInfo,
    ) -> DeploymentContext:
        """Fetch all context data needed for session creation from deployment info.

        Args:
            deployment_info: Deployment information

        Returns:
            DeploymentContext: Context data needed for session creation
        """
        return await self._db_source.fetch_deployment_context(deployment_info)

    # Auto-scaling operations

    @deployment_repository_resilience.apply()
    async def fetch_metrics_for_autoscaling(
        self,
        deployments: Sequence[DeploymentInfo],
        auto_scaling_rules: Mapping[uuid.UUID, Sequence[AutoScalingRule]],
    ) -> AutoScalingMetricsData:
        """Fetch all metrics needed for auto-scaling calculations.

        Args:
            deployments: List of deployments to fetch metrics for
            auto_scaling_rules: Auto-scaling rules by endpoint ID

        Returns:
            AutoScalingMetricsData containing all metrics needed for calculations
        """
        # Collect endpoint IDs
        endpoint_ids = {deployment.id for deployment in deployments}

        # Fetch routes for all endpoints
        routes_by_endpoint = await self._db_source.fetch_active_routes_by_endpoint_ids(endpoint_ids)

        # Determine which metrics we need to fetch based on rules
        metric_requested_sessions: list[SessionId] = []
        metric_requested_kernels: list[KernelId] = []
        metric_requested_endpoints: list[uuid.UUID] = []
        kernels_by_session_id: dict[SessionId, list[KernelId]] = defaultdict(list)

        for deployment in deployments:
            rules = auto_scaling_rules.get(deployment.id, [])
            for rule in rules:
                if rule.condition.metric_source == AutoScalingMetricSource.KERNEL:
                    # Need to fetch kernel metrics for this endpoint's sessions
                    for route in routes_by_endpoint.get(deployment.id, []):
                        if route.session_id:
                            metric_requested_sessions.append(route.session_id)
                elif rule.condition.metric_source == AutoScalingMetricSource.INFERENCE_FRAMEWORK:
                    # Need to fetch endpoint metrics
                    metric_requested_endpoints.append(deployment.id)

        # Fetch kernel data if needed
        if metric_requested_sessions:
            # Fetch kernels for sessions
            kernel_rows = await self._db_source.fetch_kernels_by_session_ids(
                list(set(metric_requested_sessions))
            )
            for kernel_id, session_id in kernel_rows:
                kernels_by_session_id[session_id].append(kernel_id)
                metric_requested_kernels.append(kernel_id)

        # Batch fetch metrics from Valkey
        kernel_statistics_by_id: dict[KernelId, Mapping[str, Any] | None] = {}
        endpoint_statistics_by_id: dict[uuid.UUID, Mapping[str, Any] | None] = {}

        if metric_requested_kernels:
            kernel_live_stats = await KernelStatistics.batch_load_by_kernel_impl(
                self._valkey_stat,
                cast(list[SessionId], metric_requested_kernels),
            )
            kernel_statistics_by_id = {
                kernel_id: metric
                for kernel_id, metric in zip(
                    metric_requested_kernels, kernel_live_stats, strict=True
                )
            }

        if metric_requested_endpoints:
            endpoint_live_stats = await EndpointStatistics.batch_load_by_endpoint_impl(
                self._valkey_stat,
                metric_requested_endpoints,
            )
            endpoint_statistics_by_id = {
                endpoint_id: metric
                for endpoint_id, metric in zip(
                    metric_requested_endpoints, endpoint_live_stats, strict=True
                )
            }

        return AutoScalingMetricsData(
            kernel_statistics=kernel_statistics_by_id,
            endpoint_statistics=endpoint_statistics_by_id,
            routes_by_endpoint=routes_by_endpoint,
            kernels_by_session=kernels_by_session_id,
        )

    @deployment_repository_resilience.apply()
    async def calculate_desired_replicas_for_deployment(
        self,
        deployment: DeploymentInfo,
        auto_scaling_rules: Sequence[AutoScalingRule],
        metrics_data: AutoScalingMetricsData,
    ) -> int | None:
        """Calculate desired replicas for a deployment based on auto-scaling rules.

        Args:
            deployment: Deployment to calculate for
            auto_scaling_rules: Auto-scaling rules to evaluate
            metrics_data: All metrics data needed for calculations

        Returns:
            Desired replica count if change is needed, None otherwise
        """
        if not auto_scaling_rules:
            return None

        current_datetime = datetime.now(UTC)
        current_replica_count = deployment.replica_spec.target_replica_count
        routes = metrics_data.routes_by_endpoint.get(deployment.id, [])

        for rule in auto_scaling_rules:
            # Calculate current metric value based on source
            current_value: Decimal | None = None
            should_trigger = False

            if rule.condition.metric_source == AutoScalingMetricSource.KERNEL:
                # Aggregate kernel metrics
                metric_aggregated_value = Decimal("0")
                metric_found_kernel_count = 0

                for route in routes:
                    if route.session_id:
                        for kernel_id in metrics_data.kernels_by_session.get(route.session_id, []):
                            kernel_stat = metrics_data.kernel_statistics.get(kernel_id)
                            if not kernel_stat:
                                continue
                            if rule.condition.metric_name not in kernel_stat:
                                continue
                            metric_found_kernel_count += 1
                            metric_value = cast(
                                dict[str, Any], kernel_stat[rule.condition.metric_name]
                            )
                            metric_aggregated_value += Decimal(str(metric_value.get("pct", 0)))

                if metric_found_kernel_count == 0:
                    log.warning(
                        "AUTOSCALE(e:{}, rule:{}): skipping - metric {} not found",
                        deployment.id,
                        rule.id,
                        rule.condition.metric_name,
                    )
                    continue

                current_value = metric_aggregated_value / Decimal(metric_found_kernel_count)

            elif rule.condition.metric_source == AutoScalingMetricSource.INFERENCE_FRAMEWORK:
                # Use endpoint metrics
                endpoint_stat = metrics_data.endpoint_statistics.get(deployment.id)
                if not endpoint_stat:
                    log.warning(
                        "AUTOSCALE(e:{}, rule:{}): skipping - no endpoint statistics",
                        deployment.id,
                        rule.id,
                    )
                    continue
                if rule.condition.metric_name not in endpoint_stat:
                    log.warning(
                        "AUTOSCALE(e:{}, rule:{}): skipping - metric {} not found",
                        deployment.id,
                        rule.id,
                        rule.condition.metric_name,
                    )
                    continue

                metric_value = cast(dict[str, Any], endpoint_stat[rule.condition.metric_name])
                route_count = len(routes) if routes else 1
                metric_type = metric_value.get("__type")
                match metric_type:
                    case "HISTOGRAM":
                        log.exception("Unable to set auto-scaling rule on histogram metrics. Skip")
                        continue
                    case "GAUGE" | "COUNTER" | _:
                        current_metric_value = metric_value.get("current", 0)
                        try:
                            current_value = Decimal(str(current_metric_value)) / Decimal(
                                route_count
                            )
                        except DecimalException:
                            log.exception(
                                "Unable parse metric value '{}' to decimal. Skip",
                                current_metric_value,
                            )
                            continue

            # Evaluate threshold comparison
            if current_value is not None:
                threshold = Decimal(rule.condition.threshold)

                if rule.condition.comparator == AutoScalingMetricComparator.LESS_THAN:
                    should_trigger = current_value < threshold
                elif rule.condition.comparator == AutoScalingMetricComparator.LESS_THAN_OR_EQUAL:
                    should_trigger = current_value <= threshold
                elif rule.condition.comparator == AutoScalingMetricComparator.GREATER_THAN:
                    should_trigger = current_value > threshold
                elif rule.condition.comparator == AutoScalingMetricComparator.GREATER_THAN_OR_EQUAL:
                    should_trigger = current_value >= threshold

                log.debug(
                    "AUTOSCALE(e:{}, rule:{}): {} {} {}: {}",
                    deployment.id,
                    rule.id,
                    current_value,
                    rule.condition.comparator.value,
                    threshold,
                    should_trigger,
                )

            if should_trigger:
                # Calculate new replica count
                new_replica_count = max(0, current_replica_count + rule.action.step_size)

                # Check min/max limits
                if (
                    rule.action.min_replicas is not None
                    and new_replica_count < rule.action.min_replicas
                ):
                    log.info(
                        "AUTOSCALE(e:{}, rule:{}): new count {} below min {}",
                        deployment.id,
                        rule.id,
                        new_replica_count,
                        rule.action.min_replicas,
                    )
                    continue

                if (
                    rule.action.max_replicas is not None
                    and new_replica_count > rule.action.max_replicas
                ):
                    log.info(
                        "AUTOSCALE(e:{}, rule:{}): new count {} above max {}",
                        deployment.id,
                        rule.id,
                        new_replica_count,
                        rule.action.max_replicas,
                    )
                    continue

                # Check cooldown period
                if rule.last_triggered_at is not None:
                    cooldown_end = rule.last_triggered_at + timedelta(
                        seconds=rule.action.cooldown_seconds
                    )
                    if current_datetime < cooldown_end:
                        log.info(
                            "AUTOSCALE(e:{}, rule:{}): in cooldown until {}",
                            deployment.id,
                            rule.id,
                            cooldown_end,
                        )
                        continue

                log.info(
                    "AUTOSCALE(e:{}, rule:{}): triggering scale from {} to {}",
                    deployment.id,
                    rule.id,
                    current_replica_count,
                    new_replica_count,
                )

                # Update last triggered time
                await self._db_source.update_autoscaling_rule_triggered(rule.id, current_datetime)

                return new_replica_count

        return None

    @deployment_repository_resilience.apply()
    async def fetch_session_statuses_by_route_ids(
        self,
        route_ids: set[uuid.UUID],
    ) -> Mapping[uuid.UUID, SessionStatus | None]:
        """Fetch session IDs for multiple routes."""
        return await self._db_source.fetch_session_statuses_by_route_ids(route_ids)

    @deployment_repository_resilience.apply()
    async def update_endpoint_route_info(
        self,
        endpoint_id: uuid.UUID,
    ) -> None:
        # Generate route connection info
        connection_info = await self._db_source.generate_route_connection_info(endpoint_id)

        # Get health check config
        health_check_config = await self._db_source.get_endpoint_health_check_config(endpoint_id)

        # Update Redis with route info
        await self._valkey_live.update_appproxy_redis_info(
            endpoint_id,
            connection_info,
            health_check_config,
        )

    @deployment_repository_resilience.apply()
    async def get_endpoint_id_by_session(
        self,
        session_id: uuid.UUID,
    ) -> uuid.UUID | None:
        """
        Get endpoint ID associated with a session.

        Args:
            session_id: ID of the session

        Returns:
            Endpoint ID if found, None otherwise
        """
        return await self._db_source.get_endpoint_id_by_session(session_id)

    @deployment_repository_resilience.apply()
    async def fetch_route_service_discovery_info(
        self,
        route_ids: set[uuid.UUID],
    ) -> list[RouteServiceDiscoveryInfo]:
        """Fetch service discovery information for routes.

        Args:
            route_ids: Set of route IDs to fetch information for

        Returns:
            List of RouteServiceDiscoveryInfo containing kernel host/port and endpoint details
        """
        return await self._db_source.fetch_route_service_discovery_info(route_ids)

    @deployment_repository_resilience.apply()
    async def get_default_architecture_from_scaling_group(
        self, scaling_group_name: str
    ) -> str | None:
        """
        Get the default (most common) architecture from active agents in a scaling group.
        Returns None if no active agents exist.
        """
        return await self._db_source.get_default_architecture_from_scaling_group(scaling_group_name)

    # ========== Deployment Revision Operations ==========

    @deployment_repository_resilience.apply()
    async def create_revision(
        self,
        creator: Creator[DeploymentRevisionRow],
    ) -> ModelRevisionData:
        """Create a new deployment revision."""
        return await self._db_source.create_revision(creator)

    @deployment_repository_resilience.apply()
    async def get_revision(
        self,
        revision_id: uuid.UUID,
    ) -> ModelRevisionData:
        """Get a deployment revision by ID.

        Raises:
            DeploymentRevisionNotFound: If the revision does not exist.
        """
        return await self._db_source.get_revision(revision_id)

    @deployment_repository_resilience.apply()
    async def get_revision_by_route_id(
        self,
        route_id: uuid.UUID,
    ) -> ModelRevisionData:
        """Get a deployment revision by route (replica) ID.

        Args:
            route_id: ID of the route (replica)

        Raises:
            RouteNotFound: If the route does not exist.
            DeploymentRevisionNotFound: If the route has no revision linked.
        """
        return await self._db_source.get_revision_by_route_id(route_id)

    @deployment_repository_resilience.apply()
    async def get_current_revision(
        self,
        endpoint_id: uuid.UUID,
    ) -> ModelRevisionData:
        """Get the current revision of a deployment.

        Args:
            endpoint_id: ID of the deployment endpoint

        Raises:
            EndpointNotFound: If the endpoint does not exist.
            DeploymentRevisionNotFound: If the endpoint has no current revision.
        """
        return await self._db_source.get_current_revision(endpoint_id)

    @deployment_repository_resilience.apply()
    async def search_revisions(
        self,
        querier: BatchQuerier,
    ) -> RevisionSearchResult:
        """Search deployment revisions with pagination and filtering."""
        return await self._db_source.search_revisions(querier)

    @deployment_repository_resilience.apply()
    async def get_latest_revision_number(
        self,
        endpoint_id: uuid.UUID,
    ) -> int | None:
        """Get the latest revision number for an endpoint.

        Returns None if no revisions exist for the endpoint.
        """
        return await self._db_source.get_latest_revision_number(endpoint_id)

    @deployment_repository_resilience.apply()
    async def update_endpoint(
        self,
        updater: Updater[EndpointRow],
    ) -> DeploymentInfo:
        """Update an endpoint using the provided updater spec.

        Returns:
            DeploymentInfo: The updated endpoint information.

        Raises:
            EndpointNotFound: If the endpoint does not exist.
        """
        return await self._db_source.update_endpoint(updater)

    @deployment_repository_resilience.apply()
    async def update_current_revision(
        self,
        endpoint_id: uuid.UUID,
        revision_id: uuid.UUID,
    ) -> uuid.UUID | None:
        """Update the current revision of a deployment.

        Returns:
            The previous revision ID, or None if there was no previous revision.
        """
        return await self._db_source.update_current_revision(endpoint_id, revision_id)

    # ========== Deployment Auto-Scaling Policy Operations ==========

    @deployment_repository_resilience.apply()
    async def create_auto_scaling_policy(
        self,
        creator: Creator[DeploymentAutoScalingPolicyRow],
    ) -> DeploymentAutoScalingPolicyData:
        """Create a new auto-scaling policy for an endpoint."""
        return await self._db_source.create_auto_scaling_policy(creator)

    @deployment_repository_resilience.apply()
    async def get_auto_scaling_policy(
        self,
        endpoint_id: uuid.UUID,
    ) -> DeploymentAutoScalingPolicyData:
        """Get the auto-scaling policy for an endpoint.

        Raises:
            AutoScalingPolicyNotFound: If no policy exists for the endpoint.
        """
        return await self._db_source.get_auto_scaling_policy(endpoint_id)

    @deployment_repository_resilience.apply()
    async def update_auto_scaling_policy(
        self,
        updater: Updater[DeploymentAutoScalingPolicyRow],
    ) -> DeploymentAutoScalingPolicyData:
        """Update an auto-scaling policy.

        Raises:
            AutoScalingPolicyNotFound: If the policy does not exist.
        """
        return await self._db_source.update_auto_scaling_policy(updater)

    @deployment_repository_resilience.apply()
    async def delete_auto_scaling_policy(
        self,
        purger: Purger[DeploymentAutoScalingPolicyRow],
    ) -> PurgerResult[DeploymentAutoScalingPolicyRow] | None:
        """Delete an auto-scaling policy by primary key.

        Returns:
            PurgerResult containing the deleted row, or None if no policy existed.
        """
        return await self._db_source.delete_auto_scaling_policy(purger)

    @deployment_repository_resilience.apply()
    async def create_deployment_policy(
        self,
        creator: Creator[DeploymentPolicyRow],
    ) -> DeploymentPolicyData:
        """Create a new deployment policy for an endpoint."""
        return await self._db_source.create_deployment_policy(creator)

    @deployment_repository_resilience.apply()
    async def get_deployment_policy(
        self,
        endpoint_id: uuid.UUID,
    ) -> DeploymentPolicyData:
        """Get the deployment policy for an endpoint.

        Raises:
            DeploymentPolicyNotFound: If no policy exists for the endpoint.
        """
        return await self._db_source.get_deployment_policy(endpoint_id)

    @deployment_repository_resilience.apply()
    async def update_deployment_policy(
        self,
        updater: Updater[DeploymentPolicyRow],
    ) -> DeploymentPolicyData:
        """Update a deployment policy.

        Raises:
            DeploymentPolicyNotFound: If the policy does not exist.
        """
        return await self._db_source.update_deployment_policy(updater)

    @deployment_repository_resilience.apply()
    async def delete_deployment_policy(
        self,
        purger: Purger[DeploymentPolicyRow],
    ) -> PurgerResult[DeploymentPolicyRow] | None:
        """Delete a deployment policy by primary key.

        Returns:
            PurgerResult containing the deleted row, or None if no policy existed.
        """
        return await self._db_source.delete_deployment_policy(purger)

    # ===================
    # Route operations
    # ===================

    @deployment_repository_resilience.apply()
    async def create_route(
        self,
        creator: Creator[RoutingRow],
    ) -> uuid.UUID:
        """Create a new route using the provided creator.

        The Creator contains a RouteCreatorSpec that defines the route properties.

        Returns:
            UUID of the newly created route.
        """
        return await self._db_source.create_route(creator)

    @deployment_repository_resilience.apply()
    async def update_route(
        self,
        updater: Updater[RoutingRow],
    ) -> bool:
        """Update a route using the provided updater.

        The Updater contains a RouteUpdaterSpec or RouteStatusUpdaterSpec
        that defines which fields to update.

        Returns:
            True if the route was updated, False if not found.
        """
        return await self._db_source.update_route(updater)

    @deployment_repository_resilience.apply()
    async def search_routes(
        self,
        querier: BatchQuerier,
    ) -> RouteSearchResult:
        """Search routes with pagination and filtering.

        Args:
            querier: BatchQuerier containing conditions, orders, and pagination

        Returns:
            RouteSearchResult with items, total_count, and pagination info
        """
        return await self._db_source.search_routes(querier)

    @deployment_repository_resilience.apply()
    async def get_route(
        self,
        route_id: uuid.UUID,
    ) -> RouteInfo | None:
        """Get a route by ID.

        Args:
            route_id: ID of the route (replica)

        Returns:
            RouteInfo if found, None otherwise
        """
        return await self._db_source.get_route(route_id)

    @deployment_repository_resilience.apply()
    async def search_endpoints(
        self,
        querier: BatchQuerier,
    ) -> DeploymentInfoSearchResult:
        """Search endpoints with pagination and filtering.

        Args:
            querier: BatchQuerier containing conditions, orders, and pagination

        Returns:
            DeploymentInfoSearchResult with items, total_count, and pagination info
        """
        return await self._db_source.search_endpoints(querier)

    # ========== Access Token Operations ==========

    @deployment_repository_resilience.apply()
    async def create_access_token(
        self,
        creator: Creator[EndpointTokenRow],
    ) -> EndpointTokenRow:
        """Create a new access token for a model deployment.

        Args:
            creator: Creator containing the EndpointTokenCreatorSpec.

        Returns:
            Created EndpointTokenRow.
        """
        return await self._db_source.create_access_token(creator)

    # ========== Additional Search Operations ==========

    @deployment_repository_resilience.apply()
    async def search_auto_scaling_rules(
        self,
        querier: BatchQuerier,
    ) -> AutoScalingRuleSearchResult:
        """Search auto-scaling rules with pagination and filtering.

        Args:
            querier: BatchQuerier containing conditions, orders, and pagination.

        Returns:
            AutoScalingRuleSearchResult with items, total_count, and pagination info.
        """
        return await self._db_source.search_auto_scaling_rules(querier)

    @deployment_repository_resilience.apply()
    async def search_access_tokens(
        self,
        querier: BatchQuerier,
    ) -> AccessTokenSearchResult:
        """Search access tokens with pagination and filtering.

        Args:
            querier: BatchQuerier containing conditions, orders, and pagination.

        Returns:
            AccessTokenSearchResult with items, total_count, and pagination info.
        """
        return await self._db_source.search_access_tokens(querier)



from __future__ import annotations

import uuid
from collections.abc import Sequence

from ai.backend.manager.data.storage_namespace.types import StorageNamespaceData
from ai.backend.manager.repositories.base import BatchQuerier, OffsetPagination
from ai.backend.manager.repositories.storage_namespace.options import StorageNamespaceConditions
from ai.backend.manager.services.storage_namespace.actions.search import (
    SearchStorageNamespacesAction,
)
from ai.backend.manager.services.storage_namespace.processors import StorageNamespaceProcessors


async def load_storage_namespaces_by_ids(
    processor: StorageNamespaceProcessors,
    namespace_ids: Sequence[uuid.UUID],
) -> list[StorageNamespaceData | None]:
    """Batch load storage namespaces by their IDs.

    Args:
        processor: The storage namespace processor.
        namespace_ids: Sequence of namespace UUIDs to load.

    Returns:
        List of StorageNamespaceData (or None if not found) in the same order as namespace_ids.
    """
    if not namespace_ids:
        return []

    querier = BatchQuerier(
        pagination=OffsetPagination(limit=len(namespace_ids)),
        conditions=[StorageNamespaceConditions.by_ids(namespace_ids)],
    )

    action_result = await processor.search_storage_namespaces.wait_for_complete(
        SearchStorageNamespacesAction(querier=querier)
    )

    namespace_map = {namespace.id: namespace for namespace in action_result.namespaces}
    return [namespace_map.get(namespace_id) for namespace_id in namespace_ids]



"""
Common calculation utilities for usage bucket metrics.

These functions provide Fair Share metric calculations used across
domain, project, and user usage bucket types.
"""

from __future__ import annotations

from datetime import date
from decimal import Decimal

from ai.backend.common.types import ResourceSlot
from ai.backend.manager.api.gql.fair_share.types import ResourceSlotGQL

SECONDS_PER_DAY = Decimal("86400")


def calculate_average_daily_usage(
    resource_usage: ResourceSlotGQL,
    period_start: date,
    period_end: date,
) -> ResourceSlotGQL:
    """
    Calculate average daily resource usage during the bucket period.

    For each resource type, computes: resource_usage / (bucket_duration_days * SECONDS_PER_DAY)

    The bucket period is inclusive on both ends: [period_start, period_end].
    For example, period_start=Feb 1, period_end=Feb 1 means 1 day.

    Args:
        resource_usage: Total resource usage in resource-seconds
        period_start: Bucket period start date (inclusive)
        period_end: Bucket period end date (inclusive)

    Returns:
        Average daily usage per resource type. Units match the resource type
        (e.g., CPU cores, memory bytes).

    Note:
        Returns empty ResourceSlotGQL if bucket duration is zero.
    """
    bucket_duration_days = Decimal((period_end - period_start).days + 1)

    if bucket_duration_days == 0:
        return ResourceSlotGQL(entries=[])

    # Convert ResourceSlotGQL to ResourceSlot for calculation
    usage_slot = ResourceSlot({
        entry.resource_type: entry.quantity for entry in resource_usage.entries
    })
    avg_daily = ResourceSlot()

    for resource_type, quantity in usage_slot.items():
        avg_daily[resource_type] = quantity / (bucket_duration_days * SECONDS_PER_DAY)

    return ResourceSlotGQL.from_resource_slot(avg_daily)


def calculate_usage_capacity_ratio(
    resource_usage: ResourceSlotGQL,
    capacity_snapshot: ResourceSlotGQL,
) -> ResourceSlotGQL:
    """
    Calculate usage ratio against total available capacity.

    For each resource type, computes: resource_usage / capacity_snapshot

    Args:
        resource_usage: Total resource usage in resource-seconds
        capacity_snapshot: Total available capacity snapshot

    Returns:
        Usage ratio in seconds. A value of 86400 means full utilization
        for one day. Values can exceed this if usage exceeds capacity.

    Note:
        Excludes resource types where capacity is zero to avoid division by zero.
    """
    # Convert ResourceSlotGQL to ResourceSlot for calculation
    usage_slot = ResourceSlot({
        entry.resource_type: entry.quantity for entry in resource_usage.entries
    })
    capacity_slot = ResourceSlot({
        entry.resource_type: entry.quantity for entry in capacity_snapshot.entries
    })
    ratio_slot = ResourceSlot()

    for resource_type, usage_quantity in usage_slot.items():
        capacity_quantity = capacity_slot.get(resource_type, Decimal(0))

        if capacity_quantity == 0:
            # Skip resources with zero capacity
            continue

        ratio_slot[resource_type] = usage_quantity / capacity_quantity

    return ResourceSlotGQL.from_resource_slot(ratio_slot)



import logging

from ai.backend.common.events.event_types.agent.anycast import AgentStartedEvent
from ai.backend.common.events.event_types.schedule.anycast import (
    DoDeploymentLifecycleEvent,
    DoDeploymentLifecycleIfNeededEvent,
    DoRouteLifecycleEvent,
    DoRouteLifecycleIfNeededEvent,
    DoSokovanProcessIfNeededEvent,
    DoSokovanProcessScheduleEvent,
)
from ai.backend.common.events.event_types.session.anycast import (
    SessionEnqueuedAnycastEvent,
    SessionTerminatedAnycastEvent,
)
from ai.backend.common.events.event_types.session.broadcast import SchedulingBroadcastEvent
from ai.backend.common.events.hub.hub import EventHub
from ai.backend.common.types import AgentId
from ai.backend.logging.utils import BraceStyleAdapter
from ai.backend.manager.scheduler.types import ScheduleType
from ai.backend.manager.sokovan.deployment.coordinator import DeploymentCoordinator
from ai.backend.manager.sokovan.deployment.route.coordinator import RouteCoordinator
from ai.backend.manager.sokovan.deployment.route.types import RouteLifecycleType
from ai.backend.manager.sokovan.deployment.types import DeploymentLifecycleType
from ai.backend.manager.sokovan.scheduler.coordinator import ScheduleCoordinator
from ai.backend.manager.sokovan.scheduling_controller import SchedulingController

log = BraceStyleAdapter(logging.getLogger(__spec__.name))


class ScheduleEventHandler:
    _schedule_coordinator: ScheduleCoordinator
    _scheduling_controller: SchedulingController
    _deployment_coordinator: DeploymentCoordinator
    _route_coordinator: RouteCoordinator
    _event_hub: EventHub

    def __init__(
        self,
        schedule_coordinator: ScheduleCoordinator,
        scheduling_controller: SchedulingController,
        deployment_coordinator: DeploymentCoordinator,
        route_coordinator: RouteCoordinator,
        event_hub: EventHub,
    ) -> None:
        self._schedule_coordinator = schedule_coordinator
        self._scheduling_controller = scheduling_controller
        self._deployment_coordinator = deployment_coordinator
        self._route_coordinator = route_coordinator
        self._event_hub = event_hub

    async def handle_session_enqueued(
        self, _context: None, _agent_id: str, _ev: SessionEnqueuedAnycastEvent
    ) -> None:
        # Request scheduling for next cycle through SchedulingController
        await self._scheduling_controller.mark_scheduling_needed([ScheduleType.SCHEDULE])

    async def handle_session_terminated(
        self, _context: None, _agent_id: str, _ev: SessionTerminatedAnycastEvent
    ) -> None:
        # Request scheduling for next cycle through SchedulingController
        await self._scheduling_controller.mark_scheduling_needed([ScheduleType.SCHEDULE])

    async def handle_agent_started(
        self, _context: None, _agent_id: str, _ev: AgentStartedEvent
    ) -> None:
        # Request scheduling for next cycle through SchedulingController
        await self._scheduling_controller.mark_scheduling_needed([ScheduleType.SCHEDULE])

    async def handle_do_sokovan_process_if_needed(
        self, _context: None, _agent_id: str, ev: DoSokovanProcessIfNeededEvent
    ) -> None:
        """Handle Sokovan process if needed event (checks marks)."""
        schedule_type = ScheduleType(ev.schedule_type)
        await self._schedule_coordinator.process_if_needed(schedule_type)

    async def handle_do_sokovan_process_schedule(
        self, _context: None, _agent_id: str, ev: DoSokovanProcessScheduleEvent
    ) -> None:
        """Handle Sokovan process schedule event (unconditional)."""
        schedule_type = ScheduleType(ev.schedule_type)
        await self._schedule_coordinator.process_schedule(schedule_type)

    async def handle_scheduling_broadcast(
        self, _context: None, _source: AgentId, ev: SchedulingBroadcastEvent
    ) -> None:
        """Handle scheduling broadcast event (individual)."""
        await self._event_hub.propagate_event(ev)

    async def handle_do_deployment_lifecycle_if_needed(
        self, _context: None, _agent_id: str, ev: DoDeploymentLifecycleIfNeededEvent
    ) -> None:
        """Handle deployment lifecycle if needed event (checks marks)."""
        lifecycle_type = DeploymentLifecycleType(ev.lifecycle_type)
        await self._deployment_coordinator.process_if_needed(lifecycle_type)

    async def handle_do_deployment_lifecycle(
        self, _context: None, _agent_id: str, ev: DoDeploymentLifecycleEvent
    ) -> None:
        """Handle deployment lifecycle event (unconditional)."""
        lifecycle_type = DeploymentLifecycleType(ev.lifecycle_type)
        await self._deployment_coordinator.process_deployment_lifecycle(lifecycle_type)

    async def handle_do_route_lifecycle_if_needed(
        self, _context: None, _agent_id: str, ev: DoRouteLifecycleIfNeededEvent
    ) -> None:
        """Handle route lifecycle if needed event (checks marks)."""
        lifecycle_type = RouteLifecycleType(ev.lifecycle_type)
        await self._route_coordinator.process_if_needed(lifecycle_type)

    async def handle_do_route_lifecycle(
        self, _context: None, _agent_id: str, ev: DoRouteLifecycleEvent
    ) -> None:
        """Handle route lifecycle event (unconditional)."""
        lifecycle_type = RouteLifecycleType(ev.lifecycle_type)
        await self._route_coordinator.process_route_lifecycle(lifecycle_type)



from dataclasses import dataclass
from typing import override

from ai.backend.manager.actions.action import BaseActionResult
from ai.backend.manager.actions.types import ActionOperationType
from ai.backend.manager.data.deployment.types import ModelDeploymentData
from ai.backend.manager.models.endpoint import EndpointRow
from ai.backend.manager.repositories.base import Updater
from ai.backend.manager.services.deployment.actions.base import DeploymentBaseAction


@dataclass
class UpdateDeploymentAction(DeploymentBaseAction):
    """Action to update an existing deployment."""

    updater: Updater[EndpointRow]

    @override
    def entity_id(self) -> str | None:
        return str(self.updater.pk_value)

    @override
    @classmethod
    def operation_type(cls) -> ActionOperationType:
        return ActionOperationType.UPDATE


@dataclass
class UpdateDeploymentActionResult(BaseActionResult):
    data: ModelDeploymentData

    @override
    def entity_id(self) -> str | None:
        return str(self.data.id)
